from __future__ import annotations

import json
import logging
import re
from collections.abc import Awaitable, Callable, Mapping, Sequence
from typing import TYPE_CHECKING, Any, cast

from pydantic import BaseModel

from memu.database.inmemory.vector import cosine_topk
from memu.prompts.retrieve.llm_category_ranker import PROMPT as LLM_CATEGORY_RANKER_PROMPT
from memu.prompts.retrieve.llm_item_ranker import PROMPT as LLM_ITEM_RANKER_PROMPT
from memu.prompts.retrieve.llm_resource_ranker import PROMPT as LLM_RESOURCE_RANKER_PROMPT
from memu.prompts.retrieve.pre_retrieval_decision import SYSTEM_PROMPT as PRE_RETRIEVAL_SYSTEM_PROMPT
from memu.prompts.retrieve.pre_retrieval_decision import USER_PROMPT as PRE_RETRIEVAL_USER_PROMPT
from memu.workflow.step import WorkflowState, WorkflowStep

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from memu.app.service import Context
    from memu.app.settings import RetrieveConfig
    from memu.database.interfaces import Database


class RetrieveMixin:
    if TYPE_CHECKING:
        retrieve_config: RetrieveConfig
        _run_workflow: Callable[..., Awaitable[WorkflowState]]
        _get_context: Callable[[], Context]
        _get_database: Callable[[], Database]
        _ensure_categories_ready: Callable[[Context, Database], Awaitable[None]]
        _get_step_llm_client: Callable[[Mapping[str, Any] | None], Any]
        _get_step_embedding_client: Callable[[Mapping[str, Any] | None], Any]
        _get_llm_client: Callable[..., Any]
        _model_dump_without_embeddings: Callable[[BaseModel], dict[str, Any]]
        _extract_json_blob: Callable[[str], str]
        _escape_prompt_value: Callable[[str], str]
        user_model: type[BaseModel]

    async def retrieve(
        self,
        queries: list[dict[str, Any]],
        where: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        if not queries:
            raise ValueError("empty_queries")
        ctx = self._get_context()
        store = self._get_database()
        original_query = self._extract_query_text(queries[-1])
        # await self._ensure_categories_ready(ctx, store)
        where_filters = self._normalize_where(where)

        context_queries_objs = queries[:-1] if len(queries) > 1 else []

        route_intention = self.retrieve_config.route_intention
        retrieve_category = self.retrieve_config.category.enabled
        retrieve_item = self.retrieve_config.item.enabled
        retrieve_resource = self.retrieve_config.resource.enabled
        sufficiency_check = self.retrieve_config.sufficiency_check

        workflow_name = "retrieve_llm" if self.retrieve_config.method == "llm" else "retrieve_rag"

        state: WorkflowState = {
            "method": self.retrieve_config.method,
            "original_query": original_query,
            "context_queries": context_queries_objs,
            "route_intention": route_intention,
            "skip_rewrite": len(queries) == 1,
            "retrieve_category": retrieve_category,
            "retrieve_item": retrieve_item,
            "retrieve_resource": retrieve_resource,
            "sufficiency_check": sufficiency_check,
            "ctx": ctx,
            "store": store,
            "where": where_filters,
        }

        result = await self._run_workflow(workflow_name, state)
        response = cast(dict[str, Any] | None, result.get("response"))
        if response is None:
            msg = "Retrieve workflow failed to produce a response"
            raise RuntimeError(msg)
        return response

    def _normalize_where(self, where: Mapping[str, Any] | None) -> dict[str, Any]:
        """Validate and clean the `where` scope filters against the configured user model."""
        if not where:
            return {}

        valid_fields = set(getattr(self.user_model, "model_fields", {}).keys())
        cleaned: dict[str, Any] = {}

        for raw_key, value in where.items():
            if value is None:
                continue
            field = raw_key.split("__", 1)[0]
            if field not in valid_fields:
                msg = f"Unknown filter field '{field}' for current user scope"
                raise ValueError(msg)
            cleaned[raw_key] = value

        return cleaned

    def _build_rag_retrieve_workflow(self) -> list[WorkflowStep]:
        steps = [
            WorkflowStep(
                step_id="route_intention",
                role="route_intention",
                handler=self._rag_route_intention,
                requires={"route_intention", "original_query", "context_queries", "skip_rewrite"},
                produces={"needs_retrieval", "rewritten_query", "active_query", "next_step_query"},
                capabilities={"llm"},
                config={"chat_llm_profile": self.retrieve_config.sufficiency_check_llm_profile},
            ),
            WorkflowStep(
                step_id="route_category",
                role="route_category",
                handler=self._rag_route_category,
                requires={"retrieve_category", "needs_retrieval", "active_query", "ctx", "store", "where"},
                produces={"category_hits", "category_summary_lookup", "query_vector"},
                capabilities={"vector"},
                config={"embed_llm_profile": "embedding"},
            ),
            WorkflowStep(
                step_id="sufficiency_after_category",
                role="sufficiency_check",
                handler=self._rag_category_sufficiency,
                requires={
                    "retrieve_category",
                    "needs_retrieval",
                    "active_query",
                    "context_queries",
                    "category_hits",
                    "ctx",
                    "store",
                    "where",
                },
                produces={"next_step_query", "proceed_to_items", "query_vector"},
                capabilities={"llm"},
                config={
                    "chat_llm_profile": self.retrieve_config.sufficiency_check_llm_profile,
                    "embed_llm_profile": "embedding",
                },
            ),
            WorkflowStep(
                step_id="recall_items",
                role="recall_items",
                handler=self._rag_recall_items,
                requires={
                    "needs_retrieval",
                    "proceed_to_items",
                    "ctx",
                    "store",
                    "where",
                    "active_query",
                    "query_vector",
                },
                produces={"item_hits", "query_vector"},
                capabilities={"vector"},
                config={"embed_llm_profile": "embedding"},
            ),
            WorkflowStep(
                step_id="sufficiency_after_items",
                role="sufficiency_check",
                handler=self._rag_item_sufficiency,
                requires={
                    "needs_retrieval",
                    "active_query",
                    "context_queries",
                    "item_hits",
                    "ctx",
                    "store",
                    "where",
                },
                produces={"next_step_query", "proceed_to_resources", "query_vector"},
                capabilities={"llm"},
                config={
                    "chat_llm_profile": self.retrieve_config.sufficiency_check_llm_profile,
                    "embed_llm_profile": "embedding",
                },
            ),
            WorkflowStep(
                step_id="recall_resources",
                role="recall_resources",
                handler=self._rag_recall_resources,
                requires={
                    "needs_retrieval",
                    "proceed_to_resources",
                    "ctx",
                    "store",
                    "where",
                    "active_query",
                    "query_vector",
                },
                produces={"resource_hits", "query_vector"},
                capabilities={"vector"},
                config={"embed_llm_profile": "embedding"},
            ),
            WorkflowStep(
                step_id="build_context",
                role="build_context",
                handler=self._rag_build_context,
                requires={"needs_retrieval", "original_query", "rewritten_query", "ctx", "store", "where"},
                produces={"response"},
                capabilities=set(),
            ),
        ]
        return steps

    def _list_retrieve_initial_keys(self) -> set[str]:
        return {
            "method",
            "original_query",
            "context_queries",
            "route_intention",
            "skip_rewrite",
            "retrieve_category",
            "retrieve_item",
            "retrieve_resource",
            "sufficiency_check",
            "ctx",
            "store",
            "where",
        }

    async def _rag_route_intention(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("route_intention"):
            state.update({
                "needs_retrieval": True,
                "rewritten_query": state["original_query"],
                "active_query": state["original_query"],
                "next_step_query": None,
                "proceed_to_items": False,
                "proceed_to_resources": False,
            })
            return state

        llm_client = self._get_step_llm_client(step_context)
        needs_retrieval, rewritten_query = await self._decide_if_retrieval_needed(
            state["original_query"],
            state["context_queries"],
            retrieved_content=None,
            llm_client=llm_client,
        )
        if state.get("skip_rewrite"):
            rewritten_query = state["original_query"]

        state.update({
            "needs_retrieval": needs_retrieval,
            "rewritten_query": rewritten_query,
            "active_query": rewritten_query,
            "next_step_query": None,
            "proceed_to_items": False,
            "proceed_to_resources": False,
        })
        return state

    async def _rag_route_category(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("retrieve_category") or not state.get("needs_retrieval"):
            state["category_hits"] = []
            state["category_summary_lookup"] = {}
            state["query_vector"] = None
            return state

        embed_client = self._get_step_embedding_client(step_context)
        store = state["store"]
        where_filters = state.get("where") or {}
        category_pool = store.memory_category_repo.list_categories(where_filters)
        qvec = (await embed_client.embed([state["active_query"]]))[0]
        hits, summary_lookup = await self._rank_categories_by_summary(
            qvec,
            self.retrieve_config.category.top_k,
            state["ctx"],
            store,
            embed_client=embed_client,
            categories=category_pool,
        )
        state.update({
            "query_vector": qvec,
            "category_hits": hits,
            "category_summary_lookup": summary_lookup,
            "category_pool": category_pool,
        })
        return state

    async def _rag_category_sufficiency(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("needs_retrieval"):
            state["proceed_to_items"] = False
            return state
        if not state.get("retrieve_category") or not state.get("sufficiency_check"):
            state["proceed_to_items"] = True
            return state

        retrieved_content = ""
        store = state["store"]
        where_filters = state.get("where") or {}
        category_pool = state.get("category_pool") or store.memory_category_repo.list_categories(where_filters)
        hits = state.get("category_hits") or []
        if hits:
            retrieved_content = self._format_category_content(
                hits,
                state.get("category_summary_lookup", {}),
                store,
                categories=category_pool,
            )

        llm_client = self._get_step_llm_client(step_context)
        needs_more, rewritten_query = await self._decide_if_retrieval_needed(
            state["active_query"],
            state["context_queries"],
            retrieved_content=retrieved_content or "No content retrieved yet.",
            llm_client=llm_client,
        )
        state["next_step_query"] = rewritten_query
        state["active_query"] = rewritten_query
        state["proceed_to_items"] = needs_more
        if needs_more:
            embed_client = self._get_step_embedding_client(step_context)
            state["query_vector"] = (await embed_client.embed([state["active_query"]]))[0]
        return state

    def _extract_referenced_item_ids(self, state: WorkflowState) -> set[str]:
        """Extract item IDs from category summary references."""
        from memu.utils.references import extract_references

        category_hits = state.get("category_hits") or []
        summary_lookup = state.get("category_summary_lookup", {})
        category_pool = state.get("category_pool") or {}
        referenced_item_ids: set[str] = set()

        for cid, _score in category_hits:
            # Get summary from lookup or category
            summary = summary_lookup.get(cid)
            if not summary:
                cat = category_pool.get(cid)
                if cat:
                    summary = cat.summary
            if summary:
                refs = extract_references(summary)
                referenced_item_ids.update(refs)

        return referenced_item_ids

    async def _rag_recall_items(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("retrieve_item") or not state.get("needs_retrieval") or not state.get("proceed_to_items"):
            state["item_hits"] = []
            return state

        store = state["store"]
        where_filters = state.get("where") or {}
        items_pool = store.memory_item_repo.list_items(where_filters)
        qvec = state.get("query_vector")
        if qvec is None:
            embed_client = self._get_step_embedding_client(step_context)
            qvec = (await embed_client.embed([state["active_query"]]))[0]
            state["query_vector"] = qvec
        state["item_hits"] = store.memory_item_repo.vector_search_items(
            qvec,
            self.retrieve_config.item.top_k,
            where=where_filters,
            ranking=self.retrieve_config.item.ranking,
            recency_decay_days=self.retrieve_config.item.recency_decay_days,
        )
        state["item_pool"] = items_pool
        return state

    async def _rag_item_sufficiency(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("needs_retrieval"):
            state["proceed_to_resources"] = False
            return state
        if not state.get("retrieve_item") or not state.get("sufficiency_check"):
            state["proceed_to_resources"] = True
            return state

        store = state["store"]
        where_filters = state.get("where") or {}
        items_pool = state.get("item_pool") or store.memory_item_repo.list_items(where_filters)
        retrieved_content = ""
        hits = state.get("item_hits") or []
        if hits:
            retrieved_content = self._format_item_content(hits, store, items=items_pool)

        llm_client = self._get_step_llm_client(step_context)
        needs_more, rewritten_query = await self._decide_if_retrieval_needed(
            state["active_query"],
            state["context_queries"],
            retrieved_content=retrieved_content or "No content retrieved yet.",
            llm_client=llm_client,
        )
        state["next_step_query"] = rewritten_query
        state["active_query"] = rewritten_query
        state["proceed_to_resources"] = needs_more
        if needs_more:
            embed_client = self._get_step_embedding_client(step_context)
            state["query_vector"] = (await embed_client.embed([state["active_query"]]))[0]
        return state

    async def _rag_recall_resources(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if (
            not state.get("needs_retrieval")
            or not state.get("retrieve_resource")
            or not state.get("proceed_to_resources")
        ):
            state["resource_hits"] = []
            return state

        store = state["store"]
        where_filters = state.get("where") or {}
        resource_pool = store.resource_repo.list_resources(where_filters)
        state["resource_pool"] = resource_pool
        corpus = self._resource_caption_corpus(store, resources=resource_pool)
        if not corpus:
            state["resource_hits"] = []
            return state

        qvec = state.get("query_vector")
        if qvec is None:
            embed_client = self._get_step_embedding_client(step_context)
            qvec = (await embed_client.embed([state["active_query"]]))[0]
            state["query_vector"] = qvec
        state["resource_hits"] = cosine_topk(qvec, corpus, k=self.retrieve_config.resource.top_k)
        return state

    def _rag_build_context(self, state: WorkflowState, _: Any) -> WorkflowState:
        response = {
            "needs_retrieval": bool(state.get("needs_retrieval")),
            "original_query": state["original_query"],
            "rewritten_query": state.get("rewritten_query", state["original_query"]),
            "next_step_query": state.get("next_step_query"),
            "categories": [],
            "items": [],
            "resources": [],
        }
        if state.get("needs_retrieval"):
            store = state["store"]
            where_filters = state.get("where") or {}
            categories_pool = state.get("category_pool") or store.memory_category_repo.list_categories(where_filters)
            items_pool = state.get("item_pool") or store.memory_item_repo.list_items(where_filters)
            resources_pool = state.get("resource_pool") or store.resource_repo.list_resources(where_filters)
            response["categories"] = self._materialize_hits(
                state.get("category_hits", []),
                categories_pool,
            )
            response["items"] = self._materialize_hits(state.get("item_hits", []), items_pool)
            response["resources"] = self._materialize_hits(
                state.get("resource_hits", []),
                resources_pool,
            )
        state["response"] = response
        return state

    def _build_llm_retrieve_workflow(self) -> list[WorkflowStep]:
        steps = [
            WorkflowStep(
                step_id="route_intention",
                role="route_intention",
                handler=self._llm_route_intention,
                requires={"original_query", "context_queries", "skip_rewrite"},
                produces={"needs_retrieval", "rewritten_query", "active_query", "next_step_query"},
                capabilities={"llm"},
                config={"llm_profile": self.retrieve_config.sufficiency_check_llm_profile},
            ),
            WorkflowStep(
                step_id="route_category",
                role="route_category",
                handler=self._llm_route_category,
                requires={"needs_retrieval", "active_query", "ctx", "store", "where"},
                produces={"category_hits"},
                capabilities={"llm"},
                config={"llm_profile": self.retrieve_config.llm_ranking_llm_profile},
            ),
            WorkflowStep(
                step_id="sufficiency_after_category",
                role="sufficiency_check",
                handler=self._llm_category_sufficiency,
                requires={"needs_retrieval", "active_query", "context_queries", "category_hits"},
                produces={"next_step_query", "proceed_to_items"},
                capabilities={"llm"},
                config={"llm_profile": self.retrieve_config.sufficiency_check_llm_profile},
            ),
            WorkflowStep(
                step_id="recall_items",
                role="recall_items",
                handler=self._llm_recall_items,
                requires={
                    "needs_retrieval",
                    "proceed_to_items",
                    "ctx",
                    "store",
                    "where",
                    "active_query",
                    "category_hits",
                },
                produces={"item_hits"},
                capabilities={"llm"},
                config={"llm_profile": self.retrieve_config.llm_ranking_llm_profile},
            ),
            WorkflowStep(
                step_id="sufficiency_after_items",
                role="sufficiency_check",
                handler=self._llm_item_sufficiency,
                requires={"needs_retrieval", "active_query", "context_queries", "item_hits"},
                produces={"next_step_query", "proceed_to_resources"},
                capabilities={"llm"},
                config={"llm_profile": self.retrieve_config.sufficiency_check_llm_profile},
            ),
            WorkflowStep(
                step_id="recall_resources",
                role="recall_resources",
                handler=self._llm_recall_resources,
                requires={
                    "needs_retrieval",
                    "proceed_to_resources",
                    "active_query",
                    "ctx",
                    "store",
                    "where",
                    "item_hits",
                    "category_hits",
                },
                produces={"resource_hits"},
                capabilities={"llm"},
                config={"llm_profile": self.retrieve_config.llm_ranking_llm_profile},
            ),
            WorkflowStep(
                step_id="build_context",
                role="build_context",
                handler=self._llm_build_context,
                requires={"needs_retrieval", "original_query", "rewritten_query"},
                produces={"response"},
                capabilities=set(),
            ),
        ]
        return steps

    async def _llm_route_intention(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("route_intention"):
            state.update({
                "needs_retrieval": True,
                "rewritten_query": state["original_query"],
                "active_query": state["original_query"],
                "next_step_query": None,
                "proceed_to_items": False,
                "proceed_to_resources": False,
            })
            return state

        llm_client = self._get_step_llm_client(step_context)
        needs_retrieval, rewritten_query = await self._decide_if_retrieval_needed(
            state["original_query"],
            state["context_queries"],
            retrieved_content=None,
            llm_client=llm_client,
        )
        if state.get("skip_rewrite"):
            rewritten_query = state["original_query"]

        state.update({
            "needs_retrieval": needs_retrieval,
            "rewritten_query": rewritten_query,
            "active_query": rewritten_query,
            "next_step_query": None,
            "proceed_to_items": False,
            "proceed_to_resources": False,
        })
        return state

    async def _llm_route_category(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("needs_retrieval"):
            state["category_hits"] = []
            return state
        llm_client = self._get_step_llm_client(step_context)
        store = state["store"]
        where_filters = state.get("where") or {}
        category_pool = store.memory_category_repo.list_categories(where_filters)
        hits = await self._llm_rank_categories(
            state["active_query"],
            self.retrieve_config.category.top_k,
            state["ctx"],
            store,
            llm_client=llm_client,
            categories=category_pool,
        )
        state["category_hits"] = hits
        state["category_pool"] = category_pool
        return state

    async def _llm_category_sufficiency(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("needs_retrieval"):
            state["proceed_to_items"] = False
            return state
        if not state.get("retrieve_category") or not state.get("sufficiency_check"):
            state["proceed_to_items"] = True
            return state

        retrieved_content = ""
        hits = state.get("category_hits") or []
        if hits:
            retrieved_content = self._format_llm_category_content(hits)

        llm_client = self._get_step_llm_client(step_context)
        needs_more, rewritten_query = await self._decide_if_retrieval_needed(
            state["active_query"],
            state["context_queries"],
            retrieved_content=retrieved_content or "No content retrieved yet.",
            llm_client=llm_client,
        )
        state["next_step_query"] = rewritten_query
        state["active_query"] = rewritten_query
        state["proceed_to_items"] = needs_more
        return state

    async def _llm_recall_items(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("needs_retrieval") or not state.get("proceed_to_items"):
            state["item_hits"] = []
            return state

        where_filters = state.get("where") or {}
        category_hits = state.get("category_hits", [])
        category_ids = [cat["id"] for cat in category_hits]
        llm_client = self._get_step_llm_client(step_context)
        store = state["store"]

        use_refs = getattr(self.retrieve_config.item, "use_category_references", False)
        ref_ids: list[str] = []
        if use_refs and category_hits:
            # Extract all ref_ids from category summaries
            from memu.utils.references import extract_references

            for cat in category_hits:
                summary = cat.get("summary") or ""
                ref_ids.extend(extract_references(summary))
        if ref_ids:
            # Query items by ref_ids
            items_pool = store.memory_item_repo.list_items_by_ref_ids(ref_ids, where_filters)
        else:
            items_pool = store.memory_item_repo.list_items(where_filters)

        relations = store.category_item_repo.list_relations(where_filters)
        category_pool = state.get("category_pool") or store.memory_category_repo.list_categories(where_filters)
        state["item_hits"] = await self._llm_rank_items(
            state["active_query"],
            self.retrieve_config.item.top_k,
            category_ids,
            state.get("category_hits", []),
            state["ctx"],
            store,
            llm_client=llm_client,
            categories=category_pool,
            items=items_pool,
            relations=relations,
        )
        state["item_pool"] = items_pool
        state["relation_pool"] = relations
        return state

    async def _llm_item_sufficiency(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("needs_retrieval"):
            state["proceed_to_resources"] = False
            return state
        if not state.get("retrieve_item") or not state.get("sufficiency_check"):
            state["proceed_to_resources"] = True
            return state

        retrieved_content = ""
        hits = state.get("item_hits") or []
        if hits:
            retrieved_content = self._format_llm_item_content(hits)

        llm_client = self._get_step_llm_client(step_context)
        needs_more, rewritten_query = await self._decide_if_retrieval_needed(
            state["active_query"],
            state["context_queries"],
            retrieved_content=retrieved_content or "No content retrieved yet.",
            llm_client=llm_client,
        )
        state["next_step_query"] = rewritten_query
        state["active_query"] = rewritten_query
        state["proceed_to_resources"] = needs_more
        return state

    async def _llm_recall_resources(self, state: WorkflowState, step_context: Any) -> WorkflowState:
        if not state.get("needs_retrieval") or not state.get("proceed_to_resources"):
            state["resource_hits"] = []
            return state

        llm_client = self._get_step_llm_client(step_context)
        store = state["store"]
        where_filters = state.get("where") or {}
        resource_pool = store.resource_repo.list_resources(where_filters)
        items_pool = state.get("item_pool") or store.memory_item_repo.list_items(where_filters)
        state["resource_hits"] = await self._llm_rank_resources(
            state["active_query"],
            self.retrieve_config.resource.top_k,
            state.get("category_hits", []),
            state.get("item_hits", []),
            state["ctx"],
            store,
            llm_client=llm_client,
            items=items_pool,
            resources=resource_pool,
        )
        state["resource_pool"] = resource_pool
        return state

    def _llm_build_context(self, state: WorkflowState, _: Any) -> WorkflowState:
        response = {
            "needs_retrieval": bool(state.get("needs_retrieval")),
            "original_query": state["original_query"],
            "rewritten_query": state.get("rewritten_query", state["original_query"]),
            "next_step_query": state.get("next_step_query"),
            "categories": [],
            "items": [],
            "resources": [],
        }
        if state.get("needs_retrieval"):
            response["categories"] = list(state.get("category_hits") or [])
            response["items"] = list(state.get("item_hits") or [])
            response["resources"] = list(state.get("resource_hits") or [])
        state["response"] = response
        return state

    async def _rank_categories_by_summary(
        self,
        query_vec: list[float],
        top_k: int,
        ctx: Context,
        store: Database,
        embed_client: Any | None = None,
        categories: Mapping[str, Any] | None = None,
    ) -> tuple[list[tuple[str, float]], dict[str, str]]:
        category_pool = categories if categories is not None else store.memory_category_repo.categories
        entries = [(cid, cat.summary) for cid, cat in category_pool.items() if cat.summary]
        if not entries:
            return [], {}
        summary_texts = [summary for _, summary in entries]
        client = embed_client or self._get_llm_client()
        summary_embeddings = await client.embed(summary_texts)
        corpus = [(cid, emb) for (cid, _), emb in zip(entries, summary_embeddings, strict=True)]
        hits = cosine_topk(query_vec, corpus, k=top_k)
        summary_lookup = dict(entries)
        return hits, summary_lookup

    async def _decide_if_retrieval_needed(
        self,
        query: str,
        context_queries: list[dict[str, Any]] | None,
        retrieved_content: str | None = None,
        system_prompt: str | None = None,
        llm_client: Any | None = None,
    ) -> tuple[bool, str]:
        """
        Decide if the query requires memory retrieval (or MORE retrieval) and rewrite it with context.

        Args:
            query: The current query string
            context_queries: List of previous query objects with role and content
            retrieved_content: Content retrieved so far (if checking for sufficiency)
            system_prompt: Optional system prompt override

        Returns:
            Tuple of (needs_retrieval: bool, rewritten_query: str)
            - needs_retrieval: True if retrieval/more retrieval is needed
            - rewritten_query: The rewritten query for the next step
        """
        history_text = self._format_query_context(context_queries)
        content_text = retrieved_content or "No content retrieved yet."

        prompt = self.retrieve_config.sufficiency_check_prompt or PRE_RETRIEVAL_USER_PROMPT
        user_prompt = prompt.format(
            query=self._escape_prompt_value(query),
            conversation_history=self._escape_prompt_value(history_text),
            retrieved_content=self._escape_prompt_value(content_text),
        )

        sys_prompt = system_prompt or PRE_RETRIEVAL_SYSTEM_PROMPT
        client = llm_client or self._get_llm_client()
        response = await client.chat(user_prompt, system_prompt=sys_prompt)
        decision = self._extract_decision(response)
        rewritten = self._extract_rewritten_query(response) or query

        return decision == "RETRIEVE", rewritten

    def _format_query_context(self, queries: list[dict[str, Any]] | None) -> str:
        """Format query context for prompts, including role information"""
        if not queries:
            return "No query context."

        lines = []
        for q in queries:
            if isinstance(q, str):
                # Backward compatibility
                lines.append(f"- {q}")
            elif isinstance(q, dict):
                role = q.get("role", "user")
                content = q.get("content")
                if isinstance(content, dict):
                    text = content.get("text", "")
                elif isinstance(content, str):
                    text = content
                else:
                    text = str(content)
                lines.append(f"- [{role}]: {text}")
            else:
                lines.append(f"- {q!s}")

        return "\n".join(lines)

    @staticmethod
    def _extract_query_text(query: dict[str, Any]) -> str:
        """
        Extract text content from query message structure.

        Args:
            query: Query in format {"role": "user", "content": {"text": "..."}}

        Returns:
            The extracted text string
        """
        if isinstance(query, str):
            # Backward compatibility: if it's already a string, return it
            return query

        if not isinstance(query, dict):
            raise TypeError("INVALID")

        content = query.get("content")
        if isinstance(content, dict):
            text = content.get("text", "")
            if not text:
                raise ValueError("EMPTY")
            return str(text)
        elif isinstance(content, str):
            # Also support {"role": "user", "content": "text"} format
            return content
        else:
            raise TypeError("INVALID")

    def _extract_decision(self, raw: str) -> str:
        """Extract RETRIEVE or NO_RETRIEVE decision from LLM response"""
        if not raw:
            return "RETRIEVE"  # Default to retrieve if uncertain

        match = re.search(r"<decision>(.*?)</decision>", raw, re.IGNORECASE | re.DOTALL)
        if match:
            decision = match.group(1).strip().upper()
            if "NO_RETRIEVE" in decision or "NO RETRIEVE" in decision:
                return "NO_RETRIEVE"
            if "RETRIEVE" in decision:
                return "RETRIEVE"

        upper = raw.strip().upper()
        if "NO_RETRIEVE" in upper or "NO RETRIEVE" in upper:
            return "NO_RETRIEVE"

        return "RETRIEVE"  # Default to retrieve

    def _extract_rewritten_query(self, raw: str) -> str | None:
        """Extract rewritten query from LLM response"""
        match = re.search(r"<rewritten_query>(.*?)</rewritten_query>", raw, re.IGNORECASE | re.DOTALL)
        if match:
            return match.group(1).strip()
        return None

    async def _embedding_based_retrieve(
        self,
        query: str,
        top_k: int,
        context_queries: list[dict[str, Any]] | None,
        ctx: Context,
        store: Database,
        llm_client: Any | None = None,
        where: Mapping[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Embedding-based retrieval with query rewriting and judging at each tier"""
        where_filters = self._normalize_where(where)
        category_pool = store.memory_category_repo.list_categories(where_filters)
        items_pool = store.memory_item_repo.list_items(where_filters)
        resource_pool = store.resource_repo.list_resources(where_filters)
        client = llm_client or self._get_llm_client()
        current_query = query
        qvec = (await client.embed([current_query]))[0]
        response: dict[str, Any] = {"resources": [], "items": [], "categories": [], "next_step_query": None}
        content_sections: list[str] = []

        # Tier 1: Categories
        cat_hits, summary_lookup = await self._rank_categories_by_summary(
            qvec,
            top_k,
            ctx,
            store,
            embed_client=client,
            categories=category_pool,
        )
        if cat_hits:
            response["categories"] = self._materialize_hits(cat_hits, category_pool)
            content_sections.append(
                self._format_category_content(cat_hits, summary_lookup, store, categories=category_pool)
            )

            needs_more, current_query = await self._decide_if_retrieval_needed(
                current_query,
                context_queries,
                retrieved_content="\n\n".join(content_sections),
                llm_client=client,
            )
            response["next_step_query"] = current_query
            if not needs_more:
                return response
            # Re-embed with rewritten query
            qvec = (await client.embed([current_query]))[0]

        # Tier 2: Items
        item_hits = store.memory_item_repo.vector_search_items(qvec, top_k, where=where_filters)
        if item_hits:
            response["items"] = self._materialize_hits(item_hits, items_pool)
            content_sections.append(self._format_item_content(item_hits, store, items=items_pool))

            needs_more, current_query = await self._decide_if_retrieval_needed(
                current_query,
                context_queries,
                retrieved_content="\n\n".join(content_sections),
                llm_client=client,
            )
            response["next_step_query"] = current_query
            if not needs_more:
                return response
            # Re-embed with rewritten query
            qvec = (await client.embed([current_query]))[0]

        # Tier 3: Resources
        resource_corpus = self._resource_caption_corpus(store, resources=resource_pool)
        if resource_corpus:
            res_hits = cosine_topk(qvec, resource_corpus, k=top_k)
            if res_hits:
                response["resources"] = self._materialize_hits(res_hits, resource_pool)
                content_sections.append(self._format_resource_content(res_hits, store, resources=resource_pool))

        return response

    def _materialize_hits(self, hits: Sequence[tuple[str, float]], pool: dict[str, Any]) -> list[dict[str, Any]]:
        out = []
        for _id, score in hits:
            obj = pool.get(_id)
            if not obj:
                continue
            data = self._model_dump_without_embeddings(obj)
            data["score"] = float(score)
            out.append(data)
        return out

    def _format_category_content(
        self,
        hits: list[tuple[str, float]],
        summaries: dict[str, str],
        store: Database,
        categories: Mapping[str, Any] | None = None,
    ) -> str:
        category_pool = categories if categories is not None else store.memory_category_repo.categories
        lines = []
        for cid, score in hits:
            cat = category_pool.get(cid)
            if not cat:
                continue
            summary = summaries.get(cid) or cat.summary or ""
            lines.append(f"Category: {cat.name}\nSummary: {summary}\nScore: {score:.3f}")
        return "\n\n".join(lines).strip()

    def _format_item_content(
        self, hits: list[tuple[str, float]], store: Database, items: Mapping[str, Any] | None = None
    ) -> str:
        item_pool = items if items is not None else store.memory_item_repo.items
        lines = []
        for iid, score in hits:
            item = item_pool.get(iid)
            if not item:
                continue
            lines.append(f"Memory Item ({item.memory_type}): {item.summary}\nScore: {score:.3f}")
        return "\n\n".join(lines).strip()

    def _format_resource_content(
        self, hits: list[tuple[str, float]], store: Database, resources: Mapping[str, Any] | None = None
    ) -> str:
        resource_pool = resources if resources is not None else store.resource_repo.resources
        lines = []
        for rid, score in hits:
            res = resource_pool.get(rid)
            if not res:
                continue
            caption = res.caption or f"Resource {res.url}"
            lines.append(f"Resource: {caption}\nScore: {score:.3f}")
        return "\n\n".join(lines).strip()

    def _resource_caption_corpus(
        self, store: Database, resources: Mapping[str, Any] | None = None
    ) -> list[tuple[str, list[float]]]:
        resource_pool = resources if resources is not None else store.resource_repo.resources
        corpus: list[tuple[str, list[float]]] = []
        for rid, res in resource_pool.items():
            if res.embedding:
                corpus.append((rid, res.embedding))
        return corpus

    def _extract_judgement(self, raw: str) -> str:
        if not raw:
            return "MORE"
        match = re.search(r"<judgement>(.*?)</judgement>", raw, re.IGNORECASE | re.DOTALL)
        if match:
            token = match.group(1).strip().upper()
            if "ENOUGH" in token:
                return "ENOUGH"
            if "MORE" in token:
                return "MORE"
        upper = raw.strip().upper()
        if "ENOUGH" in upper:
            return "ENOUGH"
        return "MORE"

    async def _llm_based_retrieve(
        self,
        query: str,
        top_k: int,
        context_queries: list[dict[str, Any]] | None,
        ctx: Context,
        store: Database,
        llm_client: Any | None = None,
        where: Mapping[str, Any] | None = None,
    ) -> dict[str, Any]:
        """
        LLM-based retrieval that uses language model to search and rank results
        in a hierarchical manner, with query rewriting and judging at each tier.

        Flow:
        1. Search categories with LLM, judge + rewrite query
        2. If needs more, search items from relevant categories, judge + rewrite
        3. If needs more, search resources related to context
        """
        where_filters = self._normalize_where(where)
        category_pool = store.memory_category_repo.list_categories(where_filters)
        items_pool = store.memory_item_repo.list_items(where_filters)
        relations = store.category_item_repo.list_relations(where_filters)
        resource_pool = store.resource_repo.list_resources(where_filters)
        current_query = query
        client = llm_client or self._get_llm_client()
        response: dict[str, Any] = {"resources": [], "items": [], "categories": [], "next_step_query": None}
        content_sections: list[str] = []

        # Tier 1: Search and rank categories
        category_hits = await self._llm_rank_categories(
            current_query,
            top_k,
            ctx,
            store,
            llm_client=client,
            categories=category_pool,
        )
        if category_hits:
            response["categories"] = category_hits
            content_sections.append(self._format_llm_category_content(category_hits))

            needs_more, current_query = await self._decide_if_retrieval_needed(
                current_query,
                context_queries,
                retrieved_content="\n\n".join(content_sections),
                llm_client=client,
            )
            response["next_step_query"] = current_query
            if not needs_more:
                return response

        # Tier 2: Search memory items from relevant categories
        relevant_category_ids = [cat["id"] for cat in category_hits]
        item_hits = await self._llm_rank_items(
            current_query,
            top_k,
            relevant_category_ids,
            category_hits,
            ctx,
            store,
            llm_client=client,
            categories=category_pool,
            items=items_pool,
            relations=relations,
        )
        if item_hits:
            response["items"] = item_hits
            content_sections.append(self._format_llm_item_content(item_hits))

            needs_more, current_query = await self._decide_if_retrieval_needed(
                current_query,
                context_queries,
                retrieved_content="\n\n".join(content_sections),
                llm_client=client,
            )
            response["next_step_query"] = current_query
            if not needs_more:
                return response

        # Tier 3: Search resources related to the context
        resource_hits = await self._llm_rank_resources(
            current_query,
            top_k,
            category_hits,
            item_hits,
            ctx,
            store,
            llm_client=client,
            items=items_pool,
            resources=resource_pool,
        )
        if resource_hits:
            response["resources"] = resource_hits
            content_sections.append(self._format_llm_resource_content(resource_hits))

        return response

    def _format_categories_for_llm(
        self,
        store: Database,
        category_ids: list[str] | None = None,
        categories: Mapping[str, Any] | None = None,
    ) -> str:
        """Format categories for LLM consumption"""
        categories_to_format = categories if categories is not None else store.memory_category_repo.categories
        if category_ids:
            categories_to_format = {cid: cat for cid, cat in categories_to_format.items() if cid in category_ids}

        if not categories_to_format:
            return "No categories available."

        lines = []
        for cid, cat in categories_to_format.items():
            lines.append(f"ID: {cid}")
            lines.append(f"Name: {cat.name}")
            if cat.description:
                lines.append(f"Description: {cat.description}")
            if cat.summary:
                lines.append(f"Summary: {cat.summary}")
            lines.append("---")

        return "\n".join(lines)

    def _format_items_for_llm(
        self,
        store: Database,
        category_ids: list[str] | None = None,
        items: Mapping[str, Any] | None = None,
        relations: Sequence[Any] | None = None,
    ) -> str:
        """Format memory items for LLM consumption, optionally filtered by category"""
        item_pool = items if items is not None else store.memory_item_repo.items
        relation_pool = relations if relations is not None else store.category_item_repo.relations
        items_to_format = []
        seen_item_ids = set()

        if category_ids:
            # Get items that belong to the specified categories
            for rel in relation_pool:
                if rel.category_id in category_ids:
                    item = item_pool.get(rel.item_id)
                    if item and item.id not in seen_item_ids:
                        items_to_format.append(item)
                        seen_item_ids.add(item.id)
        else:
            items_to_format = list(item_pool.values())

        if not items_to_format:
            return "No memory items available."

        lines = []
        for item in items_to_format:
            lines.append(f"ID: {item.id}")
            lines.append(f"Type: {item.memory_type}")
            lines.append(f"Summary: {item.summary}")
            lines.append("---")

        return "\n".join(lines)

    def _format_resources_for_llm(
        self,
        store: Database,
        item_ids: list[str] | None = None,
        items: Mapping[str, Any] | None = None,
        resources: Mapping[str, Any] | None = None,
    ) -> str:
        """Format resources for LLM consumption, optionally filtered by related items"""
        resource_pool = resources if resources is not None else store.resource_repo.resources
        item_pool = items if items is not None else store.memory_item_repo.items
        resources_to_format = []

        if item_ids:
            # Get resources that are related to the specified items
            resource_ids = {item_pool[iid].resource_id for iid in item_ids if iid in item_pool and iid is not None}
            resources_to_format = [
                resource_pool[rid] for rid in resource_ids if rid in resource_pool and rid is not None
            ]
        else:
            resources_to_format = list(resource_pool.values())

        if not resources_to_format:
            return "No resources available."

        lines = []
        for res in resources_to_format:
            lines.append(f"ID: {res.id}")
            lines.append(f"URL: {res.url}")
            lines.append(f"Modality: {res.modality}")
            if res.caption:
                lines.append(f"Caption: {res.caption}")
            lines.append("---")

        return "\n".join(lines)

    async def _llm_rank_categories(
        self,
        query: str,
        top_k: int,
        ctx: Context,
        store: Database,
        llm_client: Any | None = None,
        categories: Mapping[str, Any] | None = None,
    ) -> list[dict[str, Any]]:
        """Use LLM to rank categories based on query relevance"""
        category_pool = categories if categories is not None else store.memory_category_repo.categories
        if not category_pool:
            return []

        categories_data = self._format_categories_for_llm(store, categories=category_pool)
        prompt = LLM_CATEGORY_RANKER_PROMPT.format(
            query=self._escape_prompt_value(query),
            top_k=top_k,
            categories_data=self._escape_prompt_value(categories_data),
        )

        client = llm_client or self._get_llm_client()
        llm_response = await client.chat(prompt)
        return self._parse_llm_category_response(llm_response, store, categories=category_pool)

    async def _llm_rank_items(
        self,
        query: str,
        top_k: int,
        category_ids: list[str],
        category_hits: list[dict[str, Any]],
        ctx: Context,
        store: Database,
        llm_client: Any | None = None,
        categories: Mapping[str, Any] | None = None,
        items: Mapping[str, Any] | None = None,
        relations: Sequence[Any] | None = None,
    ) -> list[dict[str, Any]]:
        """Use LLM to rank memory items from relevant categories"""
        if not category_ids:
            print("[LLM Rank Items] No category_ids provided")
            return []

        item_pool = items if items is not None else store.memory_item_repo.items
        items_data = self._format_items_for_llm(store, category_ids, items=item_pool, relations=relations)
        if items_data == "No memory items available.":
            return []

        # Format relevant categories for context
        relevant_categories_info = "\n".join([
            f"- {cat['name']}: {cat.get('summary', cat.get('description', ''))}" for cat in category_hits
        ])

        prompt = LLM_ITEM_RANKER_PROMPT.format(
            query=self._escape_prompt_value(query),
            top_k=top_k,
            relevant_categories=self._escape_prompt_value(relevant_categories_info),
            items_data=self._escape_prompt_value(items_data),
        )

        client = llm_client or self._get_llm_client()
        llm_response = await client.chat(prompt)
        return self._parse_llm_item_response(llm_response, store, items=item_pool)

    async def _llm_rank_resources(
        self,
        query: str,
        top_k: int,
        category_hits: list[dict[str, Any]],
        item_hits: list[dict[str, Any]],
        ctx: Context,
        store: Database,
        llm_client: Any | None = None,
        items: Mapping[str, Any] | None = None,
        resources: Mapping[str, Any] | None = None,
    ) -> list[dict[str, Any]]:
        """Use LLM to rank resources related to the context"""
        # Get item IDs to filter resources
        item_ids = [item["id"] for item in item_hits]
        if not item_ids:
            return []

        item_pool = items if items is not None else store.memory_item_repo.items
        resource_pool = resources if resources is not None else store.resource_repo.resources
        resources_data = self._format_resources_for_llm(store, item_ids, items=item_pool, resources=resource_pool)
        if resources_data == "No resources available.":
            return []

        # Build context info
        context_parts = []
        if category_hits:
            context_parts.append("Relevant Categories:")
            context_parts.extend([f"- {cat['name']}" for cat in category_hits])
        if item_hits:
            context_parts.append("\nRelevant Memory Items:")
            context_parts.extend([f"- {item.get('summary', '')[:100]}..." for item in item_hits[:3]])

        context_info = "\n".join(context_parts)
        prompt = LLM_RESOURCE_RANKER_PROMPT.format(
            query=self._escape_prompt_value(query),
            top_k=top_k,
            context_info=self._escape_prompt_value(context_info),
            resources_data=self._escape_prompt_value(resources_data),
        )

        client = llm_client or self._get_llm_client()
        llm_response = await client.chat(prompt)
        return self._parse_llm_resource_response(llm_response, store, resources=resource_pool)

    def _parse_llm_category_response(
        self, raw_response: str, store: Database, categories: Mapping[str, Any] | None = None
    ) -> list[dict[str, Any]]:
        """Parse LLM category ranking response"""
        category_pool = categories if categories is not None else store.memory_category_repo.categories
        results = []
        try:
            json_blob = self._extract_json_blob(raw_response)
            parsed = json.loads(json_blob)

            if "categories" in parsed and isinstance(parsed["categories"], list):
                category_ids = parsed["categories"]
                # Return categories in the order provided by LLM (already sorted by relevance)
                for cat_id in category_ids:
                    if isinstance(cat_id, str):
                        cat = category_pool.get(cat_id)
                        if cat:
                            cat_data = self._model_dump_without_embeddings(cat)
                            results.append(cat_data)
        except Exception as e:
            logger.warning(f"Failed to parse LLM category ranking response: {e}")

        return results

    def _parse_llm_item_response(
        self, raw_response: str, store: Database, items: Mapping[str, Any] | None = None
    ) -> list[dict[str, Any]]:
        """Parse LLM item ranking response"""
        item_pool = items if items is not None else store.memory_item_repo.items
        results = []
        try:
            json_blob = self._extract_json_blob(raw_response)
            parsed = json.loads(json_blob)

            if "items" in parsed and isinstance(parsed["items"], list):
                item_ids = parsed["items"]
                # Return items in the order provided by LLM (already sorted by relevance)
                for item_id in item_ids:
                    if isinstance(item_id, str):
                        mem_item = item_pool.get(item_id)
                        if mem_item:
                            item_data = self._model_dump_without_embeddings(mem_item)
                            results.append(item_data)
        except Exception as e:
            logger.warning(f"Failed to parse LLM item ranking response: {e}")

        return results

    def _parse_llm_resource_response(
        self, raw_response: str, store: Database, resources: Mapping[str, Any] | None = None
    ) -> list[dict[str, Any]]:
        """Parse LLM resource ranking response"""
        resource_pool = resources if resources is not None else store.resource_repo.resources
        results = []
        try:
            json_blob = self._extract_json_blob(raw_response)
            parsed = json.loads(json_blob)

            if "resources" in parsed and isinstance(parsed["resources"], list):
                resource_ids = parsed["resources"]
                # Return resources in the order provided by LLM (already sorted by relevance)
                for res_id in resource_ids:
                    if isinstance(res_id, str):
                        res = resource_pool.get(res_id)
                        if res:
                            res_data = self._model_dump_without_embeddings(res)
                            results.append(res_data)
        except Exception as e:
            logger.warning(f"Failed to parse LLM resource ranking response: {e}")

        return results

    def _format_llm_category_content(self, hits: list[dict[str, Any]]) -> str:
        """Format LLM-ranked category content for judger"""
        lines = []
        for cat in hits:
            summary = cat.get("summary", "") or cat.get("description", "")
            lines.append(f"Category: {cat['name']}\nSummary: {summary}")
        return "\n\n".join(lines).strip()

    def _format_llm_item_content(self, hits: list[dict[str, Any]]) -> str:
        """Format LLM-ranked item content for judger"""
        lines = []
        for item in hits:
            lines.append(f"Memory Item ({item['memory_type']}): {item['summary']}")
        return "\n\n".join(lines).strip()

    def _format_llm_resource_content(self, hits: list[dict[str, Any]]) -> str:
        """Format LLM-ranked resource content for judger"""
        lines = []
        for res in hits:
            caption = res.get("caption", "") or f"Resource {res['url']}"
            lines.append(f"Resource: {caption}")
        return "\n\n".join(lines).strip()



import os

from memu.app import MemoryService


async def main():
    """Test with PostgreSQL storage."""
    api_key = os.environ.get("OPENAI_API_KEY")
    # Default port 5432; use 5433 if 5432 is occupied
    postgres_dsn = os.environ.get("POSTGRES_DSN", "postgresql+psycopg://postgres:postgres@127.0.0.1:5432/memu")
    file_path = os.path.abspath("tests/example/example_conversation.json")

    print("\n" + "=" * 60)
    print("[POSTGRES] Starting test...")
    print(f"[POSTGRES] DSN: {postgres_dsn}")
    print("=" * 60)

    service = MemoryService(
        llm_profiles={"default": {"api_key": api_key}},
        database_config={
            "metadata_store": {
                "provider": "postgres",
                "dsn": postgres_dsn,
                "ddl_mode": "create",
            },
            # vector_index will auto-configure to pgvector
        },
        retrieve_config={"method": "rag"},
    )

    # Memorize
    print("\n[POSTGRES] Memorizing...")
    memory = await service.memorize(resource_url=file_path, modality="conversation", user={"user_id": "123"})
    for cat in memory.get("categories", []):
        print(f"  - {cat.get('name')}: {(cat.get('summary') or '')[:80]}...")

    queries = [
        {"role": "user", "content": {"text": "Tell me about preferences"}},
        {"role": "assistant", "content": {"text": "Sure, I'll tell you about their preferences"}},
        {
            "role": "user",
            "content": {"text": "What are they"},
        },  # This is the query that will be used to retrieve the memory, the context will be used for query rewriting
    ]

    # RAG-based retrieval
    print("\n[POSTGRES] RETRIEVED - RAG")
    service.retrieve_config.method = "rag"
    result_rag = await service.retrieve(queries=queries, where={"user_id": "123"})
    print("  Categories:")
    for cat in result_rag.get("categories", [])[:3]:
        print(f"    - {cat.get('name')}: {(cat.get('summary') or cat.get('description', ''))[:80]}...")
    print("  Items:")
    for item in result_rag.get("items", [])[:3]:
        print(f"    - [{item.get('memory_type')}] {item.get('summary', '')[:100]}...")
    if result_rag.get("resources"):
        print("  Resources:")
        for res in result_rag.get("resources", [])[:3]:
            print(f"    - [{res.get('modality')}] {res.get('url', '')[:80]}...")

    # LLM-based retrieval
    print("\n[POSTGRES] RETRIEVED - LLM")
    service.retrieve_config.method = "llm"
    result_llm = await service.retrieve(queries=queries, where={"user_id": "123"})
    print("  Categories:")
    for cat in result_llm.get("categories", [])[:3]:
        print(f"    - {cat.get('name')}: {(cat.get('summary') or cat.get('description', ''))[:80]}...")
    print("  Items:")
    for item in result_llm.get("items", [])[:3]:
        print(f"    - [{item.get('memory_type')}] {item.get('summary', '')[:100]}...")
    if result_llm.get("resources"):
        print("  Resources:")
        for res in result_llm.get("resources", [])[:3]:
            print(f"    - [{res.get('modality')}] {res.get('url', '')[:80]}...")

    print("\n[POSTGRES] Test completed!")


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



from __future__ import annotations

import logging
from typing import Any

from pydantic import BaseModel

from memu.database.interfaces import Database
from memu.database.models import CategoryItem, MemoryCategory, MemoryItem, Resource
from memu.database.postgres.migration import DDLMode, run_migrations
from memu.database.postgres.repositories.category_item_repo import PostgresCategoryItemRepo
from memu.database.postgres.repositories.memory_category_repo import PostgresMemoryCategoryRepo
from memu.database.postgres.repositories.memory_item_repo import PostgresMemoryItemRepo
from memu.database.postgres.repositories.resource_repo import PostgresResourceRepo
from memu.database.postgres.schema import SQLAModels, get_sqlalchemy_models, require_sqlalchemy
from memu.database.postgres.session import SessionManager
from memu.database.repositories import CategoryItemRepo, MemoryCategoryRepo, MemoryItemRepo, ResourceRepo
from memu.database.state import DatabaseState

logger = logging.getLogger(__name__)


class PostgresStore(Database):
    resource_repo: ResourceRepo
    memory_category_repo: MemoryCategoryRepo
    memory_item_repo: MemoryItemRepo
    category_item_repo: CategoryItemRepo
    resources: dict[str, Resource]
    items: dict[str, MemoryItem]
    categories: dict[str, MemoryCategory]
    relations: list[CategoryItem]

    def __init__(
        self,
        *,
        dsn: str,
        ddl_mode: DDLMode = "create",
        vector_provider: str | None = None,
        scope_model: type[BaseModel] | None = None,
        base_model: type[BaseModel] | None = None,
        resource_model: type[Any] | None = None,
        memory_category_model: type[Any] | None = None,
        memory_item_model: type[Any] | None = None,
        category_item_model: type[Any] | None = None,
        sqla_models: SQLAModels | None = None,
    ) -> None:
        require_sqlalchemy()
        self.dsn = dsn
        self.ddl_mode = ddl_mode
        self.vector_provider = vector_provider
        self._use_vector_type = vector_provider == "pgvector"
        self._scope_model: type[BaseModel] = scope_model or base_model or BaseModel
        self._scope_fields = list(getattr(self._scope_model, "model_fields", {}).keys())
        self._state = DatabaseState()
        self._sessions = SessionManager(dsn=self.dsn)
        self._sqla_models: SQLAModels = sqla_models or get_sqlalchemy_models(scope_model=self._scope_model)
        run_migrations(dsn=self.dsn, scope_model=self._scope_model, ddl_mode=self.ddl_mode)

        resource_model = resource_model or self._sqla_models.Resource
        memory_category_model = memory_category_model or self._sqla_models.MemoryCategory
        memory_item_model = memory_item_model or self._sqla_models.MemoryItem
        category_item_model = category_item_model or self._sqla_models.CategoryItem

        self.resource_repo = PostgresResourceRepo(
            state=self._state,
            resource_model=resource_model,
            sqla_models=self._sqla_models,
            sessions=self._sessions,
            scope_fields=self._scope_fields,
        )
        self.memory_category_repo = PostgresMemoryCategoryRepo(
            state=self._state,
            memory_category_model=memory_category_model,
            sqla_models=self._sqla_models,
            sessions=self._sessions,
            scope_fields=self._scope_fields,
        )
        self.memory_item_repo = PostgresMemoryItemRepo(
            state=self._state,
            memory_item_model=memory_item_model,
            sqla_models=self._sqla_models,
            sessions=self._sessions,
            scope_fields=self._scope_fields,
            use_vector=self._use_vector_type,
        )
        self.category_item_repo = PostgresCategoryItemRepo(
            state=self._state,
            category_item_model=category_item_model,
            sqla_models=self._sqla_models,
            sessions=self._sessions,
            scope_fields=self._scope_fields,
        )

        self.resources = self._state.resources
        self.items = self._state.items
        self.categories = self._state.categories
        self.relations = self._state.relations

        # self._load_existing()

    def close(self) -> None:
        self._sessions.close()

    def _load_existing(self) -> None:
        self.resource_repo.load_existing()
        self.memory_category_repo.load_existing()
        self.memory_item_repo.load_existing()
        self.category_item_repo.load_existing()



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, Protocol, runtime_checkable

from memu.database.models import MemoryCategory


@runtime_checkable
class MemoryCategoryRepo(Protocol):
    """Repository contract for memory categories."""

    categories: dict[str, MemoryCategory]

    def list_categories(self, where: Mapping[str, Any] | None = None) -> dict[str, MemoryCategory]: ...

    def clear_categories(self, where: Mapping[str, Any] | None = None) -> dict[str, MemoryCategory]: ...

    def get_or_create_category(
        self, *, name: str, description: str, embedding: list[float], user_data: dict[str, Any]
    ) -> MemoryCategory: ...

    def update_category(
        self,
        *,
        category_id: str,
        name: str | None = None,
        description: str | None = None,
        embedding: list[float] | None = None,
        summary: str | None = None,
    ) -> MemoryCategory: ...

    def load_existing(self) -> None: ...



SYSTEM_PROMPT = """
# Task Objective
Perform two tasks:
1. **Query Rewriting** - Incorporate conversation context to make the query more specific and clear.
2. **Sufficiency Judgment** - Determine whether the retrieved content is sufficient to answer the query.

You should be conservative and only mark the result as **ENOUGH** when the retrieved content truly provides adequate information.

# Workflow
1. Review the **Query Context** to understand prior conversation and background.
2. Analyze the **Original Query**.
3. Examine the **Retrieved Content So Far**.
4. Rewrite the query by incorporating relevant context to improve clarity and specificity.
5. Judge whether the retrieved content is sufficient to answer the rewritten query based on defined criteria.

# Rules
- Query rewriting must stay faithful to the user's original intent.
- Only incorporate context that is relevant and helpful.
- Do not introduce new assumptions or external knowledge.
- Mark **ENOUGH** only if:
  - The retrieved content directly addresses the query, **and**
  - The information is specific and detailed enough, **and**
  - There are no obvious gaps or missing details.
- If any key information is missing or unclear, mark **MORE**.

# Output Format
Use the following structure:

<rewritten_query>
[Provide the rewritten query with conversation context]
</rewritten_query>

<judgement>
ENOUGH or MORE
</judgement>
"""

USER_PROMPT = """
Input:
Query Context:
{conversation_history}

Original Query:
{original_query}

Retrieved Content So Far:
{retrieved_content}
"""



from __future__ import annotations

from typing import Any


class EmbeddingBackend:
    """Defines how to talk to a specific embedding provider."""

    name: str = "base"
    embedding_endpoint: str = "/embeddings"

    def build_embedding_payload(self, *, inputs: list[str], embed_model: str) -> dict[str, Any]:
        raise NotImplementedError

    def parse_embedding_response(self, data: dict[str, Any]) -> list[list[float]]:
        raise NotImplementedError



from __future__ import annotations

import inspect
from collections.abc import Awaitable, Callable, Mapping
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from memu.workflow.interceptor import WorkflowInterceptorRegistry

WorkflowState = dict[str, Any]
WorkflowContext = Mapping[str, Any] | None
WorkflowHandler = Callable[[WorkflowState, WorkflowContext], Awaitable[WorkflowState] | WorkflowState]


@dataclass
class WorkflowStep:
    step_id: str
    role: str
    handler: WorkflowHandler
    description: str = ""
    requires: set[str] = field(default_factory=set)
    produces: set[str] = field(default_factory=set)
    capabilities: set[str] = field(default_factory=set)
    config: dict[str, Any] = field(default_factory=dict)

    def copy(self) -> WorkflowStep:
        """Create a shallow copy with copied mutable fields but shared handler."""
        return WorkflowStep(
            step_id=self.step_id,
            role=self.role,
            handler=self.handler,  # Keep reference, don't copy
            description=self.description,
            requires=set(self.requires),
            produces=set(self.produces),
            capabilities=set(self.capabilities),
            config=dict(self.config),
        )

    async def run(self, state: WorkflowState, context: WorkflowContext) -> WorkflowState:
        result = self.handler(state, context)
        if inspect.isawaitable(result):
            result = await result
        if not isinstance(result, Mapping):
            msg = f"Workflow step '{self.step_id}' must return a mapping, got {type(result).__name__}"
            raise TypeError(msg)
        return dict(result)


async def run_steps(
    name: str,
    steps: list[WorkflowStep],
    initial_state: WorkflowState,
    context: WorkflowContext = None,
    interceptor_registry: WorkflowInterceptorRegistry | None = None,
) -> WorkflowState:
    from memu.workflow.interceptor import (
        WorkflowStepContext,
        run_after_interceptors,
        run_before_interceptors,
        run_on_error_interceptors,
    )

    snapshot = interceptor_registry.snapshot() if interceptor_registry else None
    strict = interceptor_registry.strict if interceptor_registry else False

    state = dict(initial_state)
    for step in steps:
        missing = step.requires - state.keys()
        if missing:
            msg = f"Workflow '{name}' missing required keys for step '{step.step_id}': {', '.join(sorted(missing))}"
            raise KeyError(msg)
        step_context: dict[str, Any] = dict(context) if context else {}
        step_context["step_id"] = step.step_id
        if step.config:
            step_context["step_config"] = step.config

        # Build interceptor context
        interceptor_ctx = WorkflowStepContext(
            workflow_name=name,
            step_id=step.step_id,
            step_role=step.role,
            step_context=step_context,
        )

        # Run before interceptors
        if snapshot and snapshot.before:
            await run_before_interceptors(snapshot.before, interceptor_ctx, state, strict=strict)

        try:
            state = await step.run(state, step_context)
        except Exception as e:
            if snapshot and snapshot.on_error:
                await run_on_error_interceptors(snapshot.on_error, interceptor_ctx, state, e, strict=strict)
            raise

        # Run after interceptors
        if snapshot and snapshot.after:
            await run_after_interceptors(snapshot.after, interceptor_ctx, state, strict=strict)

    return state



"""Storage backends for MemU."""

from memu.database.factory import build_database
from memu.database.interfaces import (
    CategoryItemRecord,
    Database,
    MemoryCategoryRecord,
    MemoryItemRecord,
    ResourceRecord,
)
from memu.database.repositories import CategoryItemRepo, MemoryCategoryRepo, MemoryItemRepo, ResourceRepo

__all__ = [
    "CategoryItemRecord",
    "CategoryItemRepo",
    "Database",
    "MemoryCategoryRecord",
    "MemoryCategoryRepo",
    "MemoryItemRecord",
    "MemoryItemRepo",
    "ResourceRecord",
    "ResourceRepo",
    "build_database",
    "inmemory",
    "postgres",
    "schema",
    "sqlite",
]



"""
Test script to verify Nebius AI works with MemU.

Before running:
1. Get your Nebius API key from https://tokenfactory.nebius.com
2. Set environment variable: NEBIUS_API_KEY=your_key_here

Usage:
    cd memU
    set NEBIUS_API_KEY=your_key_here
    python examples/test_nebius_provider.py

Nebius provides:
- Chat models: Qwen, DeepSeek, Llama, etc.
- Embedding models: BGE, E5, Qwen3-Embedding
- All via OpenAI-compatible API
"""

import asyncio
import os
import sys

# Add src to path for local development
src_path = os.path.abspath("src")
sys.path.insert(0, src_path)

# Nebius configuration
NEBIUS_BASE_URL = "https://api.tokenfactory.nebius.com/v1/"
# Available chat models (pick one):
# - "Qwen/Qwen3-30B-A3B-Instruct-2507" (fast, cheap)
# - "Qwen/Qwen3-32B" (good balance)
# - "deepseek-ai/DeepSeek-V3-0324" (powerful)
# - "meta-llama/Llama-3.3-70B-Instruct" (reliable)
NEBIUS_CHAT_MODEL = "Qwen/Qwen3-30B-A3B-Instruct-2507"
# Available embedding models:
# - "BAAI/bge-multilingual-gemma2" (3584 dims, multilingual)
# - "BAAI/BGE-ICL" (4096 dims)
# - "intfloat/e5-mistral-7b-instruct" (4096 dims)
# - "Qwen/Qwen3-Embedding-8B" (4096 dims)
NEBIUS_EMBED_MODEL = "BAAI/bge-multilingual-gemma2"


async def test_nebius_chat():
    """Test Nebius chat completion."""
    from openai import AsyncOpenAI

    api_key = os.environ.get("NEBIUS_API_KEY")
    if not api_key:
        print("ERROR: Set NEBIUS_API_KEY environment variable")
        return False

    client = AsyncOpenAI(
        base_url=NEBIUS_BASE_URL,
        api_key=api_key,
    )

    print(f"Testing Nebius Chat API ({NEBIUS_CHAT_MODEL})...")
    try:
        response = await client.chat.completions.create(
            model=NEBIUS_CHAT_MODEL,
            messages=[
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "Say hello in one sentence."},
            ],
            max_tokens=50,
        )
        content = response.choices[0].message.content
        # Truncate long responses for display
        display = content[:100] + "..." if len(content) > 100 else content
        print(f"  Response: {display}")
        print("  ✓ Chat API works!")
        return True
    except Exception as e:
        print(f"  ✗ Chat API failed: {e}")
        return False


async def test_nebius_embeddings():
    """Test Nebius embeddings API."""
    from openai import AsyncOpenAI

    api_key = os.environ.get("NEBIUS_API_KEY")
    if not api_key:
        print("ERROR: Set NEBIUS_API_KEY environment variable")
        return False

    client = AsyncOpenAI(
        base_url=NEBIUS_BASE_URL,
        api_key=api_key,
    )

    print(f"\nTesting Nebius Embeddings API ({NEBIUS_EMBED_MODEL})...")
    try:
        response = await client.embeddings.create(
            model=NEBIUS_EMBED_MODEL,
            input=["Hello world", "This is a test"],
        )
        print(f"  Embedding dimensions: {len(response.data[0].embedding)}")
        print(f"  Number of embeddings: {len(response.data)}")
        print("  ✓ Embeddings API works!")
        return True
    except Exception as e:
        print(f"  ✗ Embeddings API failed: {e}")
        return False


async def test_memu_with_nebius():
    """Test MemU with Nebius as the LLM provider."""
    from memu.app import MemoryService

    api_key = os.environ.get("NEBIUS_API_KEY")
    if not api_key:
        print("ERROR: Set NEBIUS_API_KEY environment variable")
        return False

    print("\nTesting MemU with Nebius provider...")

    # Configure MemU to use Nebius (using dict config like example_1)
    llm_profiles = {
        "default": {
            "provider": "openai",
            "base_url": NEBIUS_BASE_URL,
            "api_key": api_key,
            "chat_model": NEBIUS_CHAT_MODEL,
            "client_backend": "sdk",
        },
        "embedding": {
            "provider": "openai",
            "base_url": NEBIUS_BASE_URL,
            "api_key": api_key,
            "embed_model": NEBIUS_EMBED_MODEL,
            "client_backend": "sdk",
        },
    }

    try:
        # Create MemU service with Nebius
        service = MemoryService(llm_profiles=llm_profiles)
        print("  ✓ MemoryService initialized with Nebius!")

        # Test memorize with a file (create temp file)
        print("\n  Testing memorize...")
        import tempfile

        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False, encoding="utf-8") as f:
            f.write("User likes Python programming and AI development. They prefer dark mode in their IDE.")
            temp_file = f.name

        try:
            result = await service.memorize(
                resource_url=temp_file,
                modality="text",
            )
            items_count = len(result.get("items", []))
            categories_count = len(result.get("categories", []))
            print(f"  ✓ Memorized! Items: {items_count}, Categories: {categories_count}")

            # Show what was extracted
            for item in result.get("items", [])[:3]:
                summary = item.get("summary", "")[:80]
                print(f"    - {summary}...")
        finally:
            os.unlink(temp_file)

        # Test retrieve
        print("\n  Testing retrieve...")
        retrieve_result = await service.retrieve(
            queries=[{"role": "user", "content": "What programming language does the user like?"}]
        )
        print(f"  ✓ Retrieved! Needs retrieval: {retrieve_result.get('needs_retrieval')}")

        items = retrieve_result.get("items", [])
        if items:
            print(f"  Found {len(items)} relevant items:")
            for item in items[:3]:
                if isinstance(item, dict):
                    summary = item.get("summary", str(item))[:60]
                else:
                    summary = str(item)[:60]
                print(f"    - {summary}...")

        print("\n" + "=" * 60)
        print("✓ SUCCESS: MemU works with Nebius!")
        print("=" * 60)
        return True

    except Exception as e:
        print(f"  ✗ MemU with Nebius failed: {e}")
        import traceback

        traceback.print_exc()
        return False


async def main():
    print("=" * 60)
    print("Nebius AI + MemU Integration Test")
    print("=" * 60)
    print(f"Base URL: {NEBIUS_BASE_URL}")
    print(f"Chat Model: {NEBIUS_CHAT_MODEL}")
    print(f"Embed Model: {NEBIUS_EMBED_MODEL}")
    print("=" * 60 + "\n")

    # Check for API key first
    if not os.environ.get("NEBIUS_API_KEY"):
        print("ERROR: NEBIUS_API_KEY environment variable not set!")
        print("\nTo get your API key:")
        print("1. Go to https://tokenfactory.nebius.com")
        print("2. Create an account / Log in")
        print("3. Get your API key")
        print("4. Run: set NEBIUS_API_KEY=your_key_here")
        return

    # Test individual APIs first
    chat_ok = await test_nebius_chat()
    embed_ok = await test_nebius_embeddings()

    if chat_ok and embed_ok:
        # Test full MemU integration
        await test_memu_with_nebius()
    else:
        print("\n" + "=" * 60)
        print("✗ FAILED: Basic API tests failed, skipping MemU test")
        print("=" * 60)


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, Protocol, runtime_checkable

from memu.database.models import CategoryItem


@runtime_checkable
class CategoryItemRepo(Protocol):
    """Repository contract for item/category relations."""

    relations: list[CategoryItem]

    def list_relations(self, where: Mapping[str, Any] | None = None) -> list[CategoryItem]: ...

    def link_item_category(self, item_id: str, cat_id: str, user_data: dict[str, Any]) -> CategoryItem: ...

    def unlink_item_category(self, item_id: str, cat_id: str) -> None: ...

    def get_item_categories(self, item_id: str) -> list[CategoryItem]: ...

    def load_existing(self) -> None: ...



from __future__ import annotations

import uuid
from collections.abc import Mapping
from typing import Any, override

from memu.database.inmemory.repositories.filter import matches_where
from memu.database.inmemory.state import InMemoryState
from memu.database.models import CategoryItem
from memu.database.repositories.category_item import CategoryItemRepo


class InMemoryCategoryItemRepository(CategoryItemRepo):
    def __init__(self, *, state: InMemoryState, category_item_model: type[CategoryItem]) -> None:
        self._state = state
        self.category_item_model = category_item_model
        self.relations: list[CategoryItem] = self._state.relations

    def list_relations(self, where: Mapping[str, Any] | None = None) -> list[CategoryItem]:
        if not where:
            return list(self.relations)
        return [rel for rel in self.relations if matches_where(rel, where)]

    def link_item_category(self, item_id: str, cat_id: str, user_data: dict[str, Any]) -> CategoryItem:
        _ = item_id  # enforced by caller via existing state
        for rel in self.relations:
            if rel.item_id == item_id and rel.category_id == cat_id:
                return rel
        rel = self.category_item_model(id=str(uuid.uuid4()), item_id=item_id, category_id=cat_id, **user_data)
        self.relations.append(rel)
        return rel

    def load_existing(self) -> None:
        return None

    @override
    def get_item_categories(self, item_id: str) -> list[CategoryItem]:
        return [rel for rel in self.relations if rel.item_id == item_id]

    @override
    def unlink_item_category(self, item_id: str, cat_id: str) -> None:
        self.relations = [rel for rel in self.relations if not (rel.item_id == item_id and rel.category_id == cat_id)]


__all__ = ["InMemoryCategoryItemRepository"]



"""SQLite memory item repository implementation."""

from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import Any

import pendulum
from sqlmodel import delete, select

from memu.database.inmemory.vector import cosine_topk, cosine_topk_salience
from memu.database.models import MemoryItem, MemoryType, compute_content_hash
from memu.database.repositories.memory_item import MemoryItemRepo
from memu.database.sqlite.repositories.base import SQLiteRepoBase
from memu.database.sqlite.schema import SQLiteSQLAModels
from memu.database.sqlite.session import SQLiteSessionManager
from memu.database.state import DatabaseState

logger = logging.getLogger(__name__)


class SQLiteMemoryItemRepo(SQLiteRepoBase, MemoryItemRepo):
    """SQLite implementation of memory item repository."""

    def __init__(
        self,
        *,
        state: DatabaseState,
        memory_item_model: type[Any],
        sqla_models: SQLiteSQLAModels,
        sessions: SQLiteSessionManager,
        scope_fields: list[str],
    ) -> None:
        """Initialize memory item repository.

        Args:
            state: Shared database state for caching.
            memory_item_model: SQLModel class for memory items.
            sqla_models: SQLAlchemy model container.
            sessions: Session manager for database connections.
            scope_fields: List of user scope field names.
        """
        super().__init__(
            state=state,
            sqla_models=sqla_models,
            sessions=sessions,
            scope_fields=scope_fields,
        )
        self._memory_item_model = memory_item_model
        self.items = self._state.items

    def get_item(self, item_id: str) -> MemoryItem | None:
        """Get a memory item by ID.

        Args:
            item_id: The item ID to look up.

        Returns:
            MemoryItem if found, None otherwise.
        """
        # Check cache first
        if item_id in self.items:
            return self.items[item_id]

        with self._sessions.session() as session:
            stmt = select(self._memory_item_model).where(self._memory_item_model.id == item_id)
            row = session.exec(stmt).first()

        if row is None:
            return None

        item = MemoryItem(
            id=row.id,
            resource_id=row.resource_id,
            memory_type=row.memory_type,
            summary=row.summary,
            embedding=self._normalize_embedding(row.embedding_json),
            created_at=row.created_at,
            updated_at=row.updated_at,
            **self._scope_kwargs_from(row),
        )
        self.items[row.id] = item
        return item

    def list_items(self, where: Mapping[str, Any] | None = None) -> dict[str, MemoryItem]:
        """List memory items matching the where clause.

        Args:
            where: Optional filter conditions.

        Returns:
            Dictionary of item ID to MemoryItem mapping.
        """
        with self._sessions.session() as session:
            stmt = select(self._memory_item_model)
            filters = self._build_filters(self._memory_item_model, where)
            if filters:
                stmt = stmt.where(*filters)
            rows = session.exec(stmt).all()

        result: dict[str, MemoryItem] = {}
        for row in rows:
            item = MemoryItem(
                id=row.id,
                resource_id=row.resource_id,
                memory_type=row.memory_type,
                summary=row.summary,
                embedding=self._normalize_embedding(row.embedding_json),
                created_at=row.created_at,
                updated_at=row.updated_at,
                **self._scope_kwargs_from(row),
            )
            result[row.id] = item
            self.items[row.id] = item

        return result

    def list_items_by_ref_ids(
        self, ref_ids: list[str], where: Mapping[str, Any] | None = None
    ) -> dict[str, MemoryItem]:
        """List items by their ref_id in the extra column.

        Args:
            ref_ids: List of ref_ids to query.
            where: Additional filter conditions.

        Returns:
            Dict mapping item_id -> MemoryItem for items whose extra.ref_id is in ref_ids.
        """
        if not ref_ids:
            return {}

        from sqlalchemy import func

        with self._sessions.session() as session:
            stmt = select(self._memory_item_model)
            filters = self._build_filters(self._memory_item_model, where)
            # Add filter for json_extract(extra, '$.ref_id') IN ref_ids (only rows with ref_id key)
            ref_id_col = func.json_extract(self._memory_item_model.extra, "$.ref_id")
            filters.append(ref_id_col.isnot(None))
            filters.append(ref_id_col.in_(ref_ids))
            if filters:
                stmt = stmt.where(*filters)
            rows = session.exec(stmt).all()

        result: dict[str, MemoryItem] = {}
        for row in rows:
            item = MemoryItem(
                id=row.id,
                resource_id=row.resource_id,
                memory_type=row.memory_type,
                summary=row.summary,
                embedding=self._normalize_embedding(row.embedding_json),
                created_at=row.created_at,
                updated_at=row.updated_at,
                **self._scope_kwargs_from(row),
            )
            result[row.id] = item
            self.items[row.id] = item

        return result

    def clear_items(self, where: Mapping[str, Any] | None = None) -> dict[str, MemoryItem]:
        """Clear items matching the where clause.

        Args:
            where: Optional filter conditions.

        Returns:
            Dictionary of deleted item ID to MemoryItem mapping.
        """
        filters = self._build_filters(self._memory_item_model, where)
        with self._sessions.session() as session:
            # First get the objects to delete
            stmt = select(self._memory_item_model)
            if filters:
                stmt = stmt.where(*filters)
            rows = session.exec(stmt).all()

            deleted: dict[str, MemoryItem] = {}
            for row in rows:
                item = MemoryItem(
                    id=row.id,
                    resource_id=row.resource_id,
                    memory_type=row.memory_type,
                    summary=row.summary,
                    embedding=self._normalize_embedding(row.embedding_json),
                    created_at=row.created_at,
                    updated_at=row.updated_at,
                    **self._scope_kwargs_from(row),
                )
                deleted[row.id] = item

            if not deleted:
                return {}

            # Delete from database
            del_stmt = delete(self._memory_item_model)
            if filters:
                del_stmt = del_stmt.where(*filters)
            session.exec(del_stmt)
            session.commit()

            # Clean up cache
            for item_id in deleted:
                self.items.pop(item_id, None)

        return deleted

    def create_item(
        self,
        *,
        resource_id: str,
        memory_type: MemoryType,
        summary: str,
        embedding: list[float],
        user_data: dict[str, Any],
        reinforce: bool = False,
        tool_record: dict[str, Any] | None = None,
    ) -> MemoryItem:
        """Create a new memory item.

        Args:
            resource_id: Associated resource ID.
            memory_type: Type of memory.
            summary: Memory summary text.
            embedding: Embedding vector.
            user_data: User scope data.
            reinforce: If True, reinforce existing item instead of creating duplicate.
            tool_record: Tool-related fields (when_to_use, metadata, tool_calls) to store in extra.

        Returns:
            Created MemoryItem object.
        """
        if reinforce and memory_type != "tool":
            return self.create_item_reinforce(
                resource_id=resource_id,
                memory_type=memory_type,
                summary=summary,
                embedding=embedding,
                user_data=user_data,
            )

        # Build extra dict with tool_record fields at top level
        extra: dict[str, Any] = {}
        if tool_record:
            if tool_record.get("when_to_use") is not None:
                extra["when_to_use"] = tool_record["when_to_use"]
            if tool_record.get("metadata") is not None:
                extra["metadata"] = tool_record["metadata"]
            if tool_record.get("tool_calls") is not None:
                extra["tool_calls"] = tool_record["tool_calls"]

        now = self._now()
        row = self._memory_item_model(
            resource_id=resource_id,
            memory_type=memory_type,
            summary=summary,
            embedding_json=self._prepare_embedding(embedding),
            extra=extra if extra else {},
            created_at=now,
            updated_at=now,
            **user_data,
        )
        with self._sessions.session() as session:
            session.add(row)
            session.commit()
            session.refresh(row)

        item = MemoryItem(
            id=row.id,
            resource_id=row.resource_id,
            memory_type=row.memory_type,
            summary=row.summary,
            embedding=embedding,
            extra=row.extra,
            created_at=row.created_at,
            updated_at=row.updated_at,
            **user_data,
        )
        self.items[row.id] = item
        return item

    def create_item_reinforce(
        self,
        *,
        resource_id: str,
        memory_type: MemoryType,
        summary: str,
        embedding: list[float],
        user_data: dict[str, Any],
    ) -> MemoryItem:
        """Create or reinforce a memory item with deduplication.

        If an item with the same content hash exists in the same scope,
        reinforce it instead of creating a duplicate.

        Args:
            resource_id: Associated resource ID.
            memory_type: Type of memory.
            summary: Memory summary text.
            embedding: Embedding vector.
            user_data: User scope data.

        Returns:
            Created or reinforced MemoryItem object.
        """
        from sqlalchemy import func

        content_hash = compute_content_hash(summary, memory_type)

        with self._sessions.session() as session:
            # Check for existing item with same hash in same scope (deduplication)
            # Use json_extract(extra, '$.content_hash') for query
            content_hash_col = func.json_extract(self._memory_item_model.extra, "$.content_hash")
            filters = [content_hash_col == content_hash]
            filters.extend(self._build_filters(self._memory_item_model, user_data))

            existing = session.exec(select(self._memory_item_model).where(*filters)).first()

            if existing:
                # Reinforce existing memory instead of creating duplicate
                current_extra = existing.extra or {}
                current_count = current_extra.get("reinforcement_count", 1)
                existing.extra = {
                    **current_extra,
                    "reinforcement_count": current_count + 1,
                    "last_reinforced_at": self._now().isoformat(),
                }
                existing.updated_at = self._now()
                session.add(existing)
                session.commit()
                session.refresh(existing)

                item = MemoryItem(
                    id=existing.id,
                    resource_id=existing.resource_id,
                    memory_type=existing.memory_type,
                    summary=existing.summary,
                    embedding=self._normalize_embedding(existing.embedding_json),
                    created_at=existing.created_at,
                    updated_at=existing.updated_at,
                    extra=existing.extra,
                    **self._scope_kwargs_from(existing),
                )
                self.items[existing.id] = item
                return item

            # Create new item with salience tracking in extra
            now = self._now()
            item_extra = user_data.pop("extra", {}) if "extra" in user_data else {}
            item_extra.update({
                "content_hash": content_hash,
                "reinforcement_count": 1,
                "last_reinforced_at": now.isoformat(),
            })

            row = self._memory_item_model(
                resource_id=resource_id,
                memory_type=memory_type,
                summary=summary,
                embedding_json=self._prepare_embedding(embedding),
                extra=item_extra,
                created_at=now,
                updated_at=now,
                **user_data,
            )

            session.add(row)
            session.commit()
            session.refresh(row)

        item = MemoryItem(
            id=row.id,
            resource_id=row.resource_id,
            memory_type=row.memory_type,
            summary=row.summary,
            embedding=embedding,
            created_at=row.created_at,
            updated_at=row.updated_at,
            extra=row.extra,
            **self._scope_kwargs_from(row),
        )
        self.items[row.id] = item
        return item

    def update_item(
        self,
        *,
        item_id: str,
        memory_type: MemoryType | None = None,
        summary: str | None = None,
        embedding: list[float] | None = None,
        extra: dict[str, Any] | None = None,
        tool_record: dict[str, Any] | None = None,
    ) -> MemoryItem:
        """Update an existing memory item.

        Args:
            item_id: ID of item to update.
            memory_type: New memory type (optional).
            summary: New summary text (optional).
            embedding: New embedding vector (optional).
            extra: Extra data to merge into existing extra dict (optional).
            tool_record: Tool-related fields (when_to_use, metadata, tool_calls) to merge into extra.

        Returns:
            Updated MemoryItem object.

        Raises:
            KeyError: If item not found.
        """
        with self._sessions.session() as session:
            stmt = select(self._memory_item_model).where(self._memory_item_model.id == item_id)
            row = session.exec(stmt).first()

            if row is None:
                msg = f"Item with id {item_id} not found"
                raise KeyError(msg)

            if memory_type is not None:
                row.memory_type = memory_type
            if summary is not None:
                row.summary = summary
            if embedding is not None:
                row.embedding_json = self._prepare_embedding(embedding)

            # Merge extra and tool_record into existing extra dict
            current_extra = row.extra or {}
            if extra is not None:
                current_extra = {**current_extra, **extra}
            if tool_record is not None:
                # Merge tool_record fields at top level
                for key in ("when_to_use", "metadata", "tool_calls"):
                    if tool_record.get(key) is not None:
                        current_extra[key] = tool_record[key]
            if extra is not None or tool_record is not None:
                row.extra = current_extra

            row.updated_at = self._now()

            session.add(row)
            session.commit()
            session.refresh(row)

        item = MemoryItem(
            id=row.id,
            resource_id=row.resource_id,
            memory_type=row.memory_type,
            summary=row.summary,
            embedding=self._normalize_embedding(row.embedding_json),
            extra=row.extra,
            created_at=row.created_at,
            updated_at=row.updated_at,
            **self._scope_kwargs_from(row),
        )
        self.items[row.id] = item
        return item

    def delete_item(self, item_id: str) -> None:
        """Delete a memory item.

        Args:
            item_id: ID of item to delete.
        """
        with self._sessions.session() as session:
            stmt = select(self._memory_item_model).where(self._memory_item_model.id == item_id)
            row = session.exec(stmt).first()
            if row:
                session.delete(row)
                session.commit()

        if item_id in self.items:
            del self.items[item_id]

    def vector_search_items(
        self,
        query_vec: list[float],
        top_k: int,
        where: Mapping[str, Any] | None = None,
        *,
        ranking: str = "similarity",
        recency_decay_days: float = 30.0,
    ) -> list[tuple[str, float]]:
        """Perform vector similarity search on memory items.

        Uses brute-force cosine similarity since SQLite doesn't have native vector support.

        Args:
            query_vec: Query embedding vector.
            top_k: Maximum number of results to return.
            where: Optional filter conditions.
            ranking: Ranking strategy - "similarity" (default) or "salience".
            recency_decay_days: Half-life for recency decay in salience ranking.

        Returns:
            List of (item_id, similarity_score) tuples.
        """
        # Load items from database with filters
        pool = self.list_items(where)

        if ranking == "salience":
            # Salience-aware ranking: similarity x reinforcement x recency
            # Read values from extra dict
            corpus = [
                (
                    i.id,
                    i.embedding,
                    (i.extra or {}).get("reinforcement_count", 1),
                    self._parse_datetime((i.extra or {}).get("last_reinforced_at")),
                )
                for i in pool.values()
            ]
            return cosine_topk_salience(query_vec, corpus, k=top_k, recency_decay_days=recency_decay_days)

        # Default: pure cosine similarity (backward compatible)
        hits = cosine_topk(query_vec, [(i.id, i.embedding) for i in pool.values()], k=top_k)
        return hits

    @staticmethod
    def _parse_datetime(dt_str: str | None) -> pendulum.DateTime | None:
        """Parse ISO datetime string from extra dict."""
        if dt_str is None:
            return None
        try:
            parsed = pendulum.parse(dt_str)
        except (ValueError, TypeError):
            return None
        else:
            if isinstance(parsed, pendulum.DateTime):
                return parsed
            return None

    def load_existing(self) -> None:
        """Load all existing items from database into cache."""
        self.list_items()


__all__ = ["SQLiteMemoryItemRepo"]



PROMPT_BLOCK_OBJECTIVE = """
# Task Objective
You are a professional Tool Memory Extractor. Your core task is to extract tool usage patterns, execution results, and learnings from agent logs or tool execution traces. This enables agents to learn from their tool usage history.
"""

PROMPT_BLOCK_WORKFLOW = """
# Workflow
Read the full resource content to understand tool execution context.
## Extract tool memories
Identify tool calls, their inputs, outputs, success/failure status, and any patterns.
## Create tool memory entries
For each significant tool usage, create a memory entry with when_to_use hints.
## Review & validate
Ensure each tool memory is actionable and helps future tool selection.
## Final output
Output Tool Memory entries.
"""

PROMPT_BLOCK_RULES = """
# Rules
## General requirements (must satisfy all)
- Each tool memory must capture: tool name, what it was used for, outcome, and when to use it again.
- Focus on patterns that help future tool selection decisions.
- Include success/failure context to help agents avoid repeated mistakes.
- Each memory should help answer: "When should I use this tool?"

## What TO Extract
- Successful tool usage patterns with context
- Failed tool attempts with lessons learned
- Tool combinations that work well together
- Performance insights (fast vs slow tools for different tasks)

## What NOT to Extract
- Trivial tool calls without learning value
- Duplicate patterns already captured
- Tool calls with no meaningful outcome

## Memory Item Content Requirements
- Include the tool name prominently
- Describe the use case or scenario
- Note the outcome (success/failure/partial)
- Provide a "when_to_use" hint for future retrieval
"""

PROMPT_BLOCK_CATEGORY = """
## Memory Categories:
{categories_str}
"""

PROMPT_BLOCK_OUTPUT = """
# Output Format (XML)
Return all memories wrapped in a single <item> element:
<item>
    <memory>
        <content>Tool memory content describing the tool usage pattern</content>
        <when_to_use>Hint for when this memory should be retrieved</when_to_use>
        <categories>
            <category>Category Name</category>
        </categories>
    </memory>
</item>
"""

PROMPT_BLOCK_EXAMPLES = """
# Examples (Input / Output / Explanation)
Example 1: Tool Memory Extraction
## Input
[Tool Call] file_reader(path="/data/config.json")
[Result] Success - Read 2048 bytes in 0.3s
[Tool Call] json_parser(content=<file_content>)
[Result] Success - Parsed config with 15 keys
[Tool Call] file_reader(path="/data/missing.json")
[Result] Error - FileNotFoundError: File does not exist
## Output
<item>
    <memory>
        <content>The file_reader tool successfully reads JSON config files from /data/ directory. Average read time is 0.3s for ~2KB files. Works well when chained with json_parser for config processing.</content>
        <when_to_use>When needing to read configuration files or JSON data from the filesystem</when_to_use>
        <categories>
            <category>file_operations</category>
        </categories>
    </memory>
    <memory>
        <content>The file_reader tool fails with FileNotFoundError when the target file doesn't exist. Should verify file existence before reading or handle the error gracefully.</content>
        <when_to_use>When handling file read errors or implementing robust file operations</when_to_use>
        <categories>
            <category>error_handling</category>
        </categories>
    </memory>
</item>
## Explanation
Two tool memories are extracted: one for successful usage pattern, one for error handling insight. Both include when_to_use hints for smart retrieval.
"""

PROMPT_BLOCK_INPUT = """
# Original Resource:
<resource>
{resource}
</resource>
"""

PROMPT = "\n\n".join([
    PROMPT_BLOCK_OBJECTIVE.strip(),
    PROMPT_BLOCK_WORKFLOW.strip(),
    PROMPT_BLOCK_RULES.strip(),
    PROMPT_BLOCK_CATEGORY.strip(),
    PROMPT_BLOCK_OUTPUT.strip(),
    PROMPT_BLOCK_EXAMPLES.strip(),
    PROMPT_BLOCK_INPUT.strip(),
])

CUSTOM_PROMPT = {
    "objective": PROMPT_BLOCK_OBJECTIVE.strip(),
    "workflow": PROMPT_BLOCK_WORKFLOW.strip(),
    "rules": PROMPT_BLOCK_RULES.strip(),
    "category": PROMPT_BLOCK_CATEGORY.strip(),
    "output": PROMPT_BLOCK_OUTPUT.strip(),
    "examples": PROMPT_BLOCK_EXAMPLES.strip(),
    "input": PROMPT_BLOCK_INPUT.strip(),
}



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from memu.database.models import MemoryCategory
from memu.database.postgres.repositories.base import PostgresRepoBase
from memu.database.postgres.session import SessionManager
from memu.database.repositories.memory_category import MemoryCategoryRepo
from memu.database.state import DatabaseState


class PostgresMemoryCategoryRepo(PostgresRepoBase, MemoryCategoryRepo):
    def __init__(
        self,
        *,
        state: DatabaseState,
        memory_category_model: type[MemoryCategory],
        sqla_models: Any,
        sessions: SessionManager,
        scope_fields: list[str],
    ) -> None:
        super().__init__(state=state, sqla_models=sqla_models, sessions=sessions, scope_fields=scope_fields)
        self._memory_category_model = memory_category_model
        self.categories: dict[str, MemoryCategory] = self._state.categories

    def list_categories(self, where: Mapping[str, Any] | None = None) -> dict[str, MemoryCategory]:
        from sqlmodel import select

        filters = self._build_filters(self._sqla_models.MemoryCategory, where)
        with self._sessions.session() as session:
            rows = session.scalars(select(self._sqla_models.MemoryCategory).where(*filters)).all()
            result: dict[str, MemoryCategory] = {}
            for row in rows:
                row.embedding = self._normalize_embedding(row.embedding)
                cat = self._cache_category(row)
                result[cat.id] = cat
        return result

    def clear_categories(self, where: Mapping[str, Any] | None = None) -> dict[str, MemoryCategory]:
        from sqlmodel import delete, select

        filters = self._build_filters(self._sqla_models.MemoryCategory, where)
        with self._sessions.session() as session:
            # First get the objects to delete
            rows = session.scalars(select(self._sqla_models.MemoryCategory).where(*filters)).all()
            deleted: dict[str, MemoryCategory] = {}
            for row in rows:
                row.embedding = self._normalize_embedding(row.embedding)
                deleted[row.id] = row

            if not deleted:
                return {}

            # Delete from database
            session.exec(delete(self._sqla_models.MemoryCategory).where(*filters))
            session.commit()

            # Clean up cache
            for cat_id in deleted:
                self.categories.pop(cat_id, None)

        return deleted

    def get_or_create_category(
        self,
        *,
        name: str,
        description: str,
        embedding: list[float],
        user_data: dict[str, Any],
    ) -> MemoryCategory:
        from sqlmodel import select

        now = self._now()
        with self._sessions.session() as session:
            filters = [self._sqla_models.MemoryCategory.name == name]
            for key, value in user_data.items():
                filters.append(getattr(self._sqla_models.MemoryCategory, key) == value)
            existing = session.scalar(select(self._sqla_models.MemoryCategory).where(*filters))

            if existing:
                updated = False
                if getattr(existing, "embedding", None) is None:
                    existing.embedding = self._prepare_embedding(embedding)
                    updated = True
                if getattr(existing, "description", None) is None:
                    existing.description = description
                    updated = True
                if updated:
                    existing.updated_at = now
                    session.add(existing)
                    session.commit()
                    session.refresh(existing)
                return self._cache_category(existing)

            cat = self._memory_category_model(
                name=name,
                description=description,
                embedding=self._prepare_embedding(embedding),
                created_at=now,
                updated_at=now,
                **user_data,
            )
            session.add(cat)
            session.commit()
            session.refresh(cat)

        return self._cache_category(cat)

    def update_category(
        self,
        *,
        category_id: str,
        name: str | None = None,
        description: str | None = None,
        embedding: list[float] | None = None,
        summary: str | None = None,
    ) -> MemoryCategory:
        from sqlmodel import select

        now = self._now()
        with self._sessions.session() as session:
            cat = session.scalar(
                select(self._sqla_models.MemoryCategory).where(self._sqla_models.MemoryCategory.id == category_id)
            )
            if cat is None:
                msg = f"Category with id {category_id} not found"
                raise KeyError(msg)

            if name is not None:
                cat.name = name
            if description is not None:
                cat.description = description
            if embedding is not None:
                cat.embedding = self._prepare_embedding(embedding)
            if summary is not None:
                cat.summary = summary

            cat.updated_at = now
            session.add(cat)
            session.commit()
            session.refresh(cat)
            cat.embedding = self._normalize_embedding(cat.embedding)

        return self._cache_category(cat)

    def load_existing(self) -> None:
        from sqlmodel import select

        with self._sessions.session() as session:
            rows = session.scalars(select(self._sqla_models.MemoryCategory)).all()
            for row in rows:
                row.embedding = self._normalize_embedding(row.embedding)
                self._cache_category(row)

    def _cache_category(self, cat: MemoryCategory) -> MemoryCategory:
        self.categories[cat.id] = cat
        return cat


__all__ = ["PostgresMemoryCategoryRepo"]



"""
Personal AI Assistant with Long-Term Memory
Powered by MemU + FastAPI on Sealos DevBox

This example demonstrates how to build a memory-enabled AI assistant
that can be deployed on Sealos DevBox with 1-click deployment.

Usage:
    # Local development
    pip install -r requirements.txt
    python main.py

    # Or with uvicorn
    uvicorn main:app --reload --host 0.0.0.0 --port 8000
"""

import os
import sys
from contextlib import asynccontextmanager

from dotenv import load_dotenv
from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel

load_dotenv()

# Add src to path for local development
src_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "src"))
if os.path.exists(src_path):
    sys.path.insert(0, src_path)

from memu.app import MemoryService

# Global memory service instance
memory_service: MemoryService | None = None


def get_llm_profiles() -> dict:
    """Build LLM profiles from environment variables."""
    base_url = os.getenv("OPENAI_BASE_URL", "https://api.openai.com/v1")
    api_key = os.getenv("OPENAI_API_KEY")

    if not api_key:
        raise ValueError("OPENAI_API_KEY environment variable is required")

    return {
        "default": {
            "provider": "openai",
            "base_url": base_url,
            "api_key": api_key,
            "chat_model": os.getenv("CHAT_MODEL", "gpt-4o-mini"),
            "client_backend": "sdk",
        },
        "embedding": {
            "provider": "openai",
            "base_url": base_url,
            "api_key": api_key,
            "embed_model": os.getenv("EMBED_MODEL", "text-embedding-3-small"),
            "client_backend": "sdk",
        },
    }


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Initialize MemU memory service on startup."""
    global memory_service

    try:
        llm_profiles = get_llm_profiles()
        memory_service = MemoryService(llm_profiles=llm_profiles)
        print("✓ MemU Memory Service initialized successfully")
    except Exception as e:
        print(f"✗ Failed to initialize MemU: {e}")
        raise

    yield

    print("Shutting down MemU Assistant...")


app = FastAPI(
    title="MemU Assistant",
    description="AI Assistant with Long-Term Memory powered by MemU",
    version="1.0.0",
    lifespan=lifespan,
)

# Enable CORS for web clients
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


# Request/Response Models
class ChatRequest(BaseModel):
    message: str
    user_id: str = "default"


class ChatResponse(BaseModel):
    response: str
    memories_used: int
    memories_stored: int


class MemorizeRequest(BaseModel):
    content: str
    user_id: str = "default"


class MemorizeResponse(BaseModel):
    status: str
    items_created: int
    categories: int


class RecallResponse(BaseModel):
    query: str
    memories_found: int
    memories: list[dict]


# API Endpoints
@app.get("/")
async def root():
    """Service information and available endpoints."""
    return {
        "service": "MemU Assistant",
        "description": "AI Assistant with Long-Term Memory",
        "version": "1.0.0",
        "status": "running",
        "endpoints": {
            "GET /": "This info",
            "GET /health": "Health check",
            "POST /chat": "Chat with memory-aware AI",
            "POST /memorize": "Store information in memory",
            "GET /recall": "Query stored memories",
        },
    }


@app.get("/health")
async def health():
    """Health check endpoint."""
    return {
        "status": "healthy",
        "memory_service_initialized": memory_service is not None,
    }


@app.post("/chat", response_model=ChatResponse)
async def chat(request: ChatRequest):
    """
    Chat with the AI assistant.

    The assistant will:
    1. Retrieve relevant memories from past conversations
    2. Use those memories as context for the response
    3. Store new information from the conversation
    """
    if not memory_service:
        raise HTTPException(status_code=503, detail="Memory service not initialized")

    try:
        # Retrieve relevant memories
        retrieve_result = await memory_service.retrieve(queries=[{"role": "user", "content": request.message}])

        memories = retrieve_result.get("items", [])

        # Build context from memories
        memory_context = []
        for mem in memories[:5]:
            if isinstance(mem, dict):
                summary = mem.get("summary", str(mem))
                memory_context.append(summary)

        # Generate response (in production, use full LLM with memory context)
        if memory_context:
            response_text = (
                f"Based on what I remember about you, here's my response to: '{request.message}'\n\n"
                f"Relevant context from our past conversations:\n"
                + "\n".join(f"- {ctx[:100]}..." if len(ctx) > 100 else f"- {ctx}" for ctx in memory_context)
            )
        else:
            response_text = f"I received your message: '{request.message}'. I don't have any relevant memories yet."

        # Store the conversation
        import tempfile

        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False, encoding="utf-8") as f:
            f.write(f"User ({request.user_id}) said: {request.message}")
            temp_file = f.name

        try:
            memorize_result = await memory_service.memorize(
                resource_url=temp_file,
                modality="text",
            )
            memories_stored = len(memorize_result.get("items", []))
        finally:
            os.unlink(temp_file)

        return ChatResponse(
            response=response_text,
            memories_used=len(memories),
            memories_stored=memories_stored,
        )

    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Chat failed: {e!s}")


@app.post("/memorize", response_model=MemorizeResponse)
async def memorize(request: MemorizeRequest):
    """Store information in long-term memory."""
    if not memory_service:
        raise HTTPException(status_code=503, detail="Memory service not initialized")

    try:
        import tempfile

        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False, encoding="utf-8") as f:
            f.write(f"[User: {request.user_id}] {request.content}")
            temp_file = f.name

        try:
            result = await memory_service.memorize(
                resource_url=temp_file,
                modality="text",
            )
            return MemorizeResponse(
                status="stored",
                items_created=len(result.get("items", [])),
                categories=len(result.get("categories", [])),
            )
        finally:
            os.unlink(temp_file)

    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Memorize failed: {e!s}")


@app.get("/recall", response_model=RecallResponse)
async def recall(query: str, limit: int = 5):
    """Recall memories related to a query."""
    if not memory_service:
        raise HTTPException(status_code=503, detail="Memory service not initialized")

    try:
        result = await memory_service.retrieve(queries=[{"role": "user", "content": query}])

        items = result.get("items", [])[:limit]
        memories = []
        for item in items:
            if isinstance(item, dict):
                memories.append({
                    "summary": item.get("summary", str(item)),
                    "category": item.get("category", "unknown"),
                })
            else:
                memories.append({"summary": str(item), "category": "unknown"})

        return RecallResponse(
            query=query,
            memories_found=len(memories),
            memories=memories,
        )

    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Recall failed: {e!s}")


if __name__ == "__main__":
    import uvicorn

    host = os.getenv("HOST", "0.0.0.0")
    port = int(os.getenv("PORT", 8000))

    print(f"Starting MemU Assistant on {host}:{port}")
    uvicorn.run("main:app", host=host, port=port, reload=True)



PROMPT = """
# Task Objective
Analyze the given video and produce two outputs:
1. A **Detailed Description** that comprehensively explains what happens in the video.
2. A **Caption** that summarizes the video in a single sentence.

# Workflow
1. Watch the video carefully from start to finish.
2. Identify the **main actions and activities** taking place over time.
3. Describe the **key objects, people, and subjects** appearing in the video.
4. Analyze the **scene, setting, and environment**.
5. Note any **audio elements**, including dialogue, narration, music, or background sounds (if available).
6. Highlight **important events or moments** in the video.
7. Describe the **temporal flow**, explaining how the events progress from beginning to end.
8. Write a concise **one-sentence caption** that captures the essence of the video.

# Rules
- Base the description strictly on observable visual and audio content from the video.
- Do not invent details that cannot be inferred from the video.
- Be comprehensive and chronological in the detailed description.
- The caption must be **exactly one sentence**.
- Use clear, neutral, and objective language.

# Output Format
Use the following structure:

<detailed_description>
[Provide the comprehensive description here]
</detailed_description>

<caption>
[Provide a one-sentence summary of the video]
</caption>

"""



import asyncio
import functools
from typing import Any, cast

import lazyllm
from lazyllm import LOG


class LazyLLMClient:
    """LAZYLLM client that relies on the LazyLLM framework."""

    DEFAULT_SOURCE = "qwen"

    def __init__(
        self,
        *,
        llm_source: str | None = None,
        vlm_source: str | None = None,
        embed_source: str | None = None,
        stt_source: str | None = None,
        chat_model: str | None = None,
        vlm_model: str | None = None,
        embed_model: str | None = None,
        stt_model: str | None = None,
    ):
        self.llm_source = llm_source or self.DEFAULT_SOURCE
        self.vlm_source = vlm_source or self.DEFAULT_SOURCE
        self.embed_source = embed_source or self.DEFAULT_SOURCE
        self.stt_source = stt_source or self.DEFAULT_SOURCE
        self.chat_model = chat_model
        self.vlm_model = vlm_model
        self.embed_model = embed_model
        self.stt_model = stt_model

    async def _call_async(self, client: Any, *args: Any, **kwargs: Any) -> Any:
        """
        Asynchronously call a LazyLLM client with given arguments and keyword arguments.
        """
        if kwargs:
            return await asyncio.to_thread(functools.partial(client, *args, **kwargs))
        else:
            return await asyncio.to_thread(client, *args)

    async def chat(
        self,
        text: str,
        *,
        max_tokens: int | None = None,
        system_prompt: str | None = None,
        temperature: float = 0.2,
    ) -> str:
        """
        Generate a summary or response for the input text using the configured LLM backend.

        Args:
            text: The input text to summarize or process.
            max_tokens: (Optional) Maximum number of tokens to generate.
            system_prompt: (Optional) System instruction to guide the LLM behavior.
        Return:
            The generated summary text as a string.
        """
        client = lazyllm.namespace("MEMU").OnlineModule(source=self.llm_source, model=self.chat_model, type="llm")
        prompt = f"{system_prompt}\n\n" if system_prompt else ""
        full_prompt = f"{prompt}text:\n{text}"
        LOG.debug(f"Summarizing text with {self.llm_source}/{self.chat_model}")
        response = await self._call_async(client, full_prompt)
        return cast(str, response)

    async def summarize(
        self,
        text: str,
        *,
        max_tokens: int | None = None,
        system_prompt: str | None = None,
    ) -> str:
        """
        Generate a summary or response for the input text using the configured LLM backend.

        Args:
            text: The input text to summarize or process.
            max_tokens: (Optional) Maximum number of tokens to generate.
            system_prompt: (Optional) System instruction to guide the LLM behavior.
        Return:
            The generated summary text as a string.
        """
        client = lazyllm.namespace("MEMU").OnlineModule(source=self.llm_source, model=self.chat_model, type="llm")
        prompt = system_prompt or "Summarize the text in one short paragraph."
        full_prompt = f"{prompt}\n\ntext:\n{text}"
        LOG.debug(f"Summarizing text with {self.llm_source}/{self.chat_model}")
        response = await self._call_async(client, full_prompt)
        return cast(str, response)

    async def vision(
        self,
        prompt: str,
        image_path: str,
        *,
        max_tokens: int | None = None,
        system_prompt: str | None = None,
    ) -> tuple[str, Any]:
        """
        Process an image with a text prompt using the configured VLM (Vision-Language Model).


        Args:
            prompt: Text prompt describing the request or question about the image.
            image_path: Path to the image file to be analyzed.
            max_tokens: (Optional) Maximum number of tokens to generate.
            system_prompt: (Optional) System instruction to guide the VLM behavior.
        Return:
            A tuple containing the generated text response and None (reserved for metadata).
        """
        client = lazyllm.namespace("MEMU").OnlineModule(source=self.vlm_source, model=self.vlm_model, type="vlm")
        LOG.debug(f"Processing image with {self.vlm_source}/{self.vlm_model}: {image_path}")
        # LazyLLM VLM accepts prompt as first positional argument and image_path as keyword argument
        response = await self._call_async(client, prompt, lazyllm_files=image_path)
        return response, None

    async def embed(
        self,
        texts: list[str],
        batch_size: int = 10,
    ) -> list[list[float]]:
        """
        Generate vector embeddings for a list of text strings.

        Args:
            texts: List of text strings to embed.
            batch_size: (Optional) Batch size for processing embeddings (default: 10).
        Return:
            A list of embedding vectors (list of floats), one for each input text.
        """
        client = lazyllm.namespace("MEMU").OnlineModule(
            source=self.embed_source, model=self.embed_model, type="embed", batch_size=batch_size
        )
        LOG.debug(f"embed {len(texts)} texts with {self.embed_source}/{self.embed_model}")
        response = await self._call_async(client, texts)
        return cast(list[list[float]], response)

    async def transcribe(
        self,
        audio_path: str,
        language: str | None = None,
        prompt: str | None = None,
    ) -> str:
        """
        Transcribe audio content to text using the configured STT (Speech-to-Text) backend.

        Args:
            audio_path: Path to the audio file to transcribe.
            language: (Optional) Language code of the audio content.
            prompt: (Optional) Text prompt to guide the transcription or translation.
        Return:
            The transcribed text as a string.
        """
        client = lazyllm.namespace("MEMU").OnlineModule(source=self.stt_source, model=self.stt_model, type="stt")
        LOG.debug(f"Transcribing audio with {self.stt_source}/{self.stt_model}: {audio_path}")
        response = await self._call_async(client, audio_path)
        return cast(str, response)



PROMPT = """
# Task Objective
Judge whether the retrieved content is sufficient to answer the user's query.

# Workflow
1. Analyze the **Query** to understand what the user is asking.
2. Review the **Retrieved Content** carefully.
3. Evaluate the retrieved content against the following criteria:
   - Does it directly address the user's question?
   - Is the information specific and detailed enough?
   - Are there obvious gaps or missing details?
   - Did the user explicitly ask to recall or remember more information?
4. Based on this evaluation, decide whether the information is sufficient or more is needed.

# Rules
- Base your judgement **only** on the provided query and retrieved content.
- Do not assume or add external knowledge.
- Do not provide additional explanations beyond the required sections.
- The final judgement must be **one word only**.

# Output Format
Use the following structure:

<consideration>
Explain your reasoning for how you made the judgement.
</consideration>

<judgement>
ENOUGH or MORE
</judgement>


# Input
Query:
{query}

Retrieved Content:
{content}
"""



"""Video processing utilities for frame extraction."""

from __future__ import annotations

import logging
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import ClassVar

logger = logging.getLogger(__name__)


class VideoFrameExtractor:
    """Extract frames from video files using ffmpeg."""

    FFMPEG_BINARIES: ClassVar[set[str]] = {"ffmpeg", "ffprobe"}

    @classmethod
    def is_ffmpeg_available(cls) -> bool:
        """Check if ffmpeg is available in the system."""
        try:
            result = cls._run_ffmpeg_command(["ffmpeg", "-version"], timeout=5, check=False)
        except (FileNotFoundError, subprocess.TimeoutExpired, ValueError):
            return False
        else:
            return result.returncode == 0

    @staticmethod
    def extract_middle_frame(video_path: str, output_path: str | None = None) -> str:
        """
        Extract the middle frame from a video file.

        Args:
            video_path: Path to the video file
            output_path: Optional output path for the frame. If None, creates a temp file.

        Returns:
            Path to the extracted frame image

        Raises:
            RuntimeError: If ffmpeg is not available or extraction fails
        """
        if not VideoFrameExtractor.is_ffmpeg_available():
            msg = "ffmpeg is not available. Please install ffmpeg to process videos."
            raise RuntimeError(msg)

        video_path_obj = VideoFrameExtractor._resolve_existing_path(video_path, description="Video file")
        safe_video_path = str(video_path_obj)

        # Create output path if not provided
        created_temp_file = False
        if output_path is None:
            with tempfile.NamedTemporaryFile(suffix=".jpg", delete=False) as tmp_file:
                output_path = tmp_file.name
            created_temp_file = True
        output_path_obj = VideoFrameExtractor._resolve_output_path(output_path)
        safe_output_path = str(output_path_obj)

        try:
            # Get video duration
            duration_cmd = [
                "ffprobe",
                "-v",
                "error",
                "-show_entries",
                "format=duration",
                "-of",
                "default=noprint_wrappers=1:nokey=1",
                safe_video_path,
            ]

            logger.debug(f"Getting video duration: {' '.join(duration_cmd)}")
            duration_result = VideoFrameExtractor._run_ffmpeg_command(duration_cmd, timeout=30)

            duration = float(duration_result.stdout.strip())
            middle_time = duration / 2

            logger.debug(f"Video duration: {duration}s, extracting frame at {middle_time}s")

            # Extract frame at middle timestamp
            extract_cmd = [
                "ffmpeg",
                "-ss",
                str(middle_time),
                "-i",
                safe_video_path,
                "-vframes",
                "1",
                "-q:v",
                "2",  # High quality
                "-y",  # Overwrite output file
                safe_output_path,
            ]

            logger.debug(f"Extracting frame: {' '.join(extract_cmd)}")
            VideoFrameExtractor._run_ffmpeg_command(extract_cmd, timeout=30)

            if not output_path_obj.exists():
                msg = f"Frame extraction failed: output file not created at {output_path_obj}"
                raise RuntimeError(msg)
            else:
                logger.info(f"Successfully extracted frame to: {output_path_obj}")
                return str(output_path_obj)

        except subprocess.CalledProcessError as e:
            if created_temp_file and output_path_obj.exists():
                output_path_obj.unlink()
            msg = f"ffmpeg/ffprobe failed: {e.stderr}"
            logger.exception(msg)
            raise RuntimeError(msg) from e
        except subprocess.TimeoutExpired as e:
            if created_temp_file and output_path_obj.exists():
                output_path_obj.unlink()
            msg = "Video processing timed out"
            logger.exception(msg)
            raise RuntimeError(msg) from e

    @staticmethod
    def extract_multiple_frames(
        video_path: str,
        num_frames: int = 3,
        output_dir: str | None = None,
    ) -> list[str]:
        """
        Extract multiple evenly-spaced frames from a video.

        Args:
            video_path: Path to the video file
            num_frames: Number of frames to extract
            output_dir: Optional output directory. If None, creates a temp directory.

        Returns:
            List of paths to extracted frame images

        Raises:
            RuntimeError: If ffmpeg is not available or extraction fails
        """
        if not VideoFrameExtractor.is_ffmpeg_available():
            msg = "ffmpeg is not available. Please install ffmpeg to process videos."
            raise RuntimeError(msg)

        video_path_obj = VideoFrameExtractor._resolve_existing_path(video_path, description="Video file")
        safe_video_path = str(video_path_obj)

        # Create output directory if not provided
        created_temp_dir = False
        if output_dir is None:
            output_dir = tempfile.mkdtemp()
            created_temp_dir = True

        output_dir_obj = VideoFrameExtractor._ensure_safe_cli_path(Path(output_dir))
        output_dir_obj.mkdir(parents=True, exist_ok=True)

        try:
            # Get video duration
            duration_cmd = [
                "ffprobe",
                "-v",
                "error",
                "-show_entries",
                "format=duration",
                "-of",
                "default=noprint_wrappers=1:nokey=1",
                safe_video_path,
            ]

            logger.debug(f"Getting video duration: {' '.join(duration_cmd)}")
            duration_result = VideoFrameExtractor._run_ffmpeg_command(duration_cmd, timeout=30)

            duration = float(duration_result.stdout.strip())

            # Calculate timestamps for evenly-spaced frames
            timestamps = [duration * (i + 1) / (num_frames + 1) for i in range(num_frames)]

            logger.debug(f"Video duration: {duration}s, extracting frames at: {timestamps}")

            frame_paths = []
            for idx, timestamp in enumerate(timestamps):
                output_path_obj = VideoFrameExtractor._resolve_output_path(str(output_dir_obj / f"frame_{idx:03d}.jpg"))

                # Extract frame at timestamp
                extract_cmd = [
                    "ffmpeg",
                    "-ss",
                    str(timestamp),
                    "-i",
                    safe_video_path,
                    "-vframes",
                    "1",
                    "-q:v",
                    "2",  # High quality
                    "-y",  # Overwrite output file
                    str(output_path_obj),
                ]

                logger.debug(f"Extracting frame {idx + 1}/{num_frames}: {' '.join(extract_cmd)}")
                VideoFrameExtractor._run_ffmpeg_command(extract_cmd, timeout=30)

                if not output_path_obj.exists():
                    msg = f"Frame extraction failed: output file not created at {output_path_obj}"
                    raise RuntimeError(msg)

                frame_paths.append(str(output_path_obj))

            logger.info(f"Successfully extracted {len(frame_paths)} frames to: {output_dir_obj}")
        except subprocess.CalledProcessError as e:
            if created_temp_dir and output_dir_obj.exists():
                shutil.rmtree(output_dir_obj)
            msg = f"ffmpeg/ffprobe failed: {e.stderr}"
            logger.exception(msg)
            raise RuntimeError(msg) from e
        except subprocess.TimeoutExpired as e:
            if created_temp_dir and output_dir_obj.exists():
                shutil.rmtree(output_dir_obj)
            msg = "Video processing timed out"
            logger.exception(msg)
            raise RuntimeError(msg) from e
        else:
            return frame_paths

    @staticmethod
    def _ensure_safe_cli_path(path_obj: Path) -> Path:
        """Ensure the given path is safe to pass to a CLI command."""
        resolved = path_obj.resolve()
        if resolved.name.startswith("-"):
            msg = f"Unsafe path for CLI execution: {resolved}"
            raise ValueError(msg)
        return resolved

    @classmethod
    def _resolve_existing_path(cls, path: str, *, description: str) -> Path:
        """Resolve and validate an existing filesystem path."""
        path_obj = Path(path)
        if not path_obj.exists():
            msg = f"{description} not found: {path}"
            raise FileNotFoundError(msg)
        return cls._ensure_safe_cli_path(path_obj)

    @classmethod
    def _resolve_output_path(cls, path: str) -> Path:
        """Resolve output paths (which may not yet exist) for CLI safety."""
        return cls._ensure_safe_cli_path(Path(path))

    @classmethod
    def _run_ffmpeg_command(
        cls,
        cmd: list[str],
        *,
        timeout: int,
        check: bool = True,
        capture_output: bool = True,
    ) -> subprocess.CompletedProcess[str]:
        """Run an ffmpeg/ffprobe command after validating the executable."""
        if not cmd:
            msg = "FFmpeg command cannot be empty."
            raise ValueError(msg)
        executable = cmd[0]
        binary_name = Path(executable).name
        if binary_name not in cls.FFMPEG_BINARIES:
            msg = f"Unsupported executable '{executable}'"
            raise ValueError(msg)
        safe_cmd = [executable, *[str(arg) for arg in cmd[1:]]]
        return subprocess.run(  # noqa: S603
            safe_cmd,
            capture_output=capture_output,
            text=True,
            timeout=timeout,
            check=check,
        )



"""
Tests for memory item reference functionality (Issue #202).

Tests cover:
1. Reference extraction from text
2. Reference stripping for clean display
3. Reference formatting as citations
4. Reference-aware category summary generation
5. Reference-aware retrieval
"""

from __future__ import annotations

from memu.utils.references import (
    build_item_reference_map,
    extract_references,
    format_references_as_citations,
    strip_references,
)


class TestExtractReferences:
    """Tests for extract_references function."""

    def test_extract_single_reference(self):
        """Should extract a single reference ID."""
        text = "User loves coffee [ref:abc123]."
        refs = extract_references(text)
        assert refs == ["abc123"]

    def test_extract_multiple_references(self):
        """Should extract multiple reference IDs in order."""
        text = "User loves coffee [ref:abc123]. Also tea [ref:def456]."
        refs = extract_references(text)
        assert refs == ["abc123", "def456"]

    def test_extract_comma_separated_references(self):
        """Should handle comma-separated IDs in single reference."""
        text = "User prefers hot drinks [ref:abc,def,ghi]."
        refs = extract_references(text)
        assert refs == ["abc", "def", "ghi"]

    def test_extract_no_duplicates(self):
        """Should not return duplicate IDs."""
        text = "Coffee [ref:abc]. More coffee [ref:abc]. Tea [ref:def]."
        refs = extract_references(text)
        assert refs == ["abc", "def"]

    def test_extract_empty_text(self):
        """Should return empty list for empty text."""
        assert extract_references("") == []
        assert extract_references(None) == []

    def test_extract_no_references(self):
        """Should return empty list when no references present."""
        text = "User loves coffee and tea."
        refs = extract_references(text)
        assert refs == []

    def test_extract_with_hyphens_and_underscores(self):
        """Should handle IDs with hyphens and underscores."""
        text = "Info [ref:item_abc-123]."
        refs = extract_references(text)
        assert refs == ["item_abc-123"]


class TestStripReferences:
    """Tests for strip_references function."""

    def test_strip_single_reference(self):
        """Should remove single reference."""
        text = "User loves coffee [ref:abc123]."
        result = strip_references(text)
        assert result == "User loves coffee."

    def test_strip_multiple_references(self):
        """Should remove all references."""
        text = "Coffee [ref:abc]. Tea [ref:def]."
        result = strip_references(text)
        assert result == "Coffee. Tea."

    def test_strip_comma_separated(self):
        """Should remove comma-separated references."""
        text = "Drinks [ref:abc,def,ghi]."
        result = strip_references(text)
        assert result == "Drinks."

    def test_strip_empty_text(self):
        """Should handle empty text."""
        assert strip_references("") == ""
        assert strip_references(None) is None

    def test_strip_no_references(self):
        """Should return text unchanged if no references."""
        text = "User loves coffee."
        result = strip_references(text)
        assert result == "User loves coffee."


class TestFormatReferencesAsCitations:
    """Tests for format_references_as_citations function."""

    def test_format_single_citation(self):
        """Should convert single reference to numbered citation."""
        text = "User loves coffee [ref:abc]."
        result = format_references_as_citations(text)
        assert result is not None
        assert "[1]" in result
        assert "[ref:abc]" not in result
        assert "References:" in result
        assert "[1] abc" in result

    def test_format_multiple_citations(self):
        """Should number citations in order of appearance."""
        text = "Coffee [ref:abc]. Tea [ref:def]."
        result = format_references_as_citations(text)
        assert result is not None
        assert "[1]" in result
        assert "[2]" in result
        assert "[1] abc" in result
        assert "[2] def" in result

    def test_format_empty_text(self):
        """Should handle empty text."""
        assert format_references_as_citations("") == ""
        assert format_references_as_citations(None) is None

    def test_format_no_references(self):
        """Should return text unchanged if no references."""
        text = "User loves coffee."
        result = format_references_as_citations(text)
        assert result == text


class TestBuildItemReferenceMap:
    """Tests for build_item_reference_map function."""

    def test_build_map_single_item(self):
        """Should format single item reference."""
        items = [("abc123", "User loves coffee")]
        result = build_item_reference_map(items)
        assert "Available memory items for reference:" in result
        assert "[ref:abc123]" in result
        assert "User loves coffee" in result

    def test_build_map_multiple_items(self):
        """Should format multiple item references."""
        items = [
            ("abc", "User loves coffee"),
            ("def", "User prefers tea"),
        ]
        result = build_item_reference_map(items)
        assert "[ref:abc]" in result
        assert "[ref:def]" in result

    def test_build_map_truncates_long_summaries(self):
        """Should truncate summaries longer than 100 chars."""
        long_summary = "x" * 150
        items = [("abc", long_summary)]
        result = build_item_reference_map(items)
        assert "..." in result
        assert len(result.split("\n")[1]) < 150

    def test_build_map_empty_list(self):
        """Should return empty string for empty list."""
        assert build_item_reference_map([]) == ""


class TestReferenceIntegration:
    """Integration tests for reference functionality."""

    def test_roundtrip_extract_and_strip(self):
        """Extracting then stripping should give clean text."""
        original = "User loves coffee [ref:abc]. Tea [ref:def]."
        refs = extract_references(original)
        clean = strip_references(original)

        assert refs == ["abc", "def"]
        assert clean is not None
        assert "[ref:" not in clean
        assert "coffee" in clean
        assert "Tea" in clean

    def test_citation_preserves_content(self):
        """Citation formatting should preserve text content."""
        original = "User loves coffee [ref:abc]."
        result = format_references_as_citations(original)

        assert result is not None
        assert "User loves coffee" in result
        assert "abc" in result  # ID should be in references section



"""
Category summary prompt with inline references to memory items.

This prompt instructs the LLM to include [ref:ITEM_ID] citations
when summarizing category content, linking statements to their
source memory items.
"""

PROMPT_BLOCK_OBJECTIVE = """
# Task Objective
You are a professional User Profile Synchronization Specialist. Your core objective is to accurately merge newly extracted user information items into the user's initial profile using only two operations: add and update.

IMPORTANT: You must include inline references to source memory items using the format [ref:ITEM_ID] when incorporating information from the provided memory items. This creates a traceable link between summary statements and their sources.
"""

PROMPT_BLOCK_WORKFLOW = """
# Workflow
## Step 1: Preprocessing & Parsing
- Input sources:
  - User Initial Profile: structured, categorized, confirmed long-term user information.
  - Newly Extracted User Information Items: each item has an ID that MUST be referenced.
- Structure parsing:
  - Initial profile: extract categories and core content; preserve original wording style and format.
  - New items: note the item ID for each piece of information to include as [ref:ID].

## Step 2: Core Operations (Update / Add)
A. Update
- When updating existing information with new data, add the reference: "User is 30 years old [ref:item_abc123]"
- If multiple items support the same fact, include multiple refs: [ref:id1,id2]

B. Add
- When adding new information, always include the source reference
- Format: "User enjoys hiking on weekends [ref:item_xyz789]"

## Step 3: Merge & Formatting
- Structured ordering: present content by category order; omit empty categories.
- Formatting rules: strictly use Markdown (# for main title, ## for category titles).
- References: ensure every new or updated fact has at least one [ref:ITEM_ID] citation.

## Step 4: Summarize
Target length: {target_length}
- Summarize the updated user markdown profile to the target length.
- PRESERVE all [ref:ITEM_ID] citations in the summary.
- Use Markdown hierarchy.

## Step 5: Output
- Output only the updated user markdown profile with inline references.
- Use Markdown hierarchy.
- Do not include explanations, operation traces, or meta text.
"""

PROMPT_BLOCK_RULES = """
# Reference Rules
1. Every piece of information from new memory items MUST have a [ref:ITEM_ID] citation
2. Use the exact item ID provided in the input
3. Place references immediately after the relevant statement
4. Multiple sources can be cited: [ref:id1,id2]
5. Existing information without new updates does not need references
"""

PROMPT_BLOCK_OUTPUT = """
# Output Format (Markdown with References)
```markdown
# {category}
## <category name>
- User information item [ref:ITEM_ID]
- User information item [ref:ITEM_ID]
## <category name>
- User information item [ref:ITEM_ID,ITEM_ID2]
```

# Critical
- Always ensure that your output does not exceed {target_length} tokens.
- ALWAYS include [ref:ITEM_ID] for information from new memory items.
- You may merge or omit unimportant information to meet the length limit.
"""

PROMPT_BLOCK_EXAMPLES = """
# Examples (Input / Output)

Topic:
Personal Basic Information

Original content:
<content>
# Personal Basic Information
## Basic Information
- The user is 28 years old
- The user currently lives in Beijing
</content>

New memory items with IDs:
<items>
- [item_a1b2c3] The user is 30 years old
- [item_d4e5f6] The user currently lives in Shanghai
- [item_g7h8i9] The user prefers Sichuan-style spicy food
</items>

Output:
# Personal Basic Information
## Basic Information
- The user is 30 years old [ref:item_a1b2c3]
- The user currently lives in Shanghai [ref:item_d4e5f6]
## Basic Preferences
- The user prefers Sichuan-style spicy food [ref:item_g7h8i9]
"""

PROMPT_BLOCK_INPUT = """
# Input
Topic:
{category}

Original content:
<content>
{original_content}
</content>

New memory items with IDs:
<items>
{new_memory_items_text}
</items>
"""

PROMPT = "\n\n".join([
    PROMPT_BLOCK_OBJECTIVE.strip(),
    PROMPT_BLOCK_WORKFLOW.strip(),
    PROMPT_BLOCK_RULES.strip(),
    PROMPT_BLOCK_OUTPUT.strip(),
    PROMPT_BLOCK_EXAMPLES.strip(),
    PROMPT_BLOCK_INPUT.strip(),
])

CUSTOM_PROMPT = {
    "objective": PROMPT_BLOCK_OBJECTIVE.strip(),
    "workflow": PROMPT_BLOCK_WORKFLOW.strip(),
    "rules": PROMPT_BLOCK_RULES.strip(),
    "output": PROMPT_BLOCK_OUTPUT.strip(),
    "examples": PROMPT_BLOCK_EXAMPLES.strip(),
    "input": PROMPT_BLOCK_INPUT.strip(),
}



from typing import Any

import aiohttp
from claude_agent_sdk import create_sdk_mcp_server, tool

BASE_URL = "https://api.memu.so"
API_KEY = "your memu api key"
USER_ID = "claude_user"
AGENT_ID = "claude_agent"


@tool("memu_memory", "Retrieve memory based on a query", {"query": str})
async def get_memory(args: dict[str, Any]) -> dict[str, Any]:
    """Retrieve memory from the memory API based on the provided query."""
    query = args["query"]
    url = f"{BASE_URL}/api/v3/memory/retrieve"
    headers = {"Authorization": f"Bearer {API_KEY}"}
    data = {"user_id": USER_ID, "agent_id": AGENT_ID, "query": query}

    async with aiohttp.ClientSession() as session, session.post(url, headers=headers, json=data) as response:
        result = await response.json()

    return {"content": [{"type": "text", "text": str(result)}]}


async def _get_todos() -> str:
    url = f"{BASE_URL}/api/v3/memory/categories"
    headers = {"Authorization": f"Bearer {API_KEY}"}
    data = {
        "user_id": USER_ID,
        "agent_id": AGENT_ID,
    }
    async with aiohttp.ClientSession() as session, session.post(url, headers=headers, json=data) as response:
        result = await response.json()

    categories = result["categories"]
    todos = ""
    for category in categories:
        if category["name"] == "todo":
            todos = category["summary"]
    return todos


@tool("memu_todos", "Retrieve todos for the user", {})
async def get_todos() -> dict[str, Any]:
    """Retrieve todos from the memory API."""
    todos = await _get_todos()
    return {"content": [{"type": "text", "text": str(todos)}]}


# Create the MCP server with the tool
memu_server = create_sdk_mcp_server(name="memu", version="1.0.0", tools=[get_memory, get_todos])



"""Demo script for MemU LangGraph Integration."""

import asyncio
import logging
import os
import sys

# Try imports and fail proactively if missing
try:
    import langgraph  # noqa: F401
    from langchain_core.tools import BaseTool

    from memu.app.service import MemoryService
    from memu.integrations.langgraph import MemULangGraphTools
except ImportError:
    print("Missing dependencies. Please run: uv sync --extra langgraph")
    sys.exit(1)

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("langgraph_demo")


async def initialize_infrastructure() -> MemULangGraphTools:
    """Initialize the MemoryService and the LangGraph adapter."""
    # Ensure OPENAI_API_KEY is present
    if not os.environ.get("OPENAI_API_KEY"):
        logger.warning("OPENAI_API_KEY not found in environment variables.")

    # In a real scenario, you might load config from file or env
    service = MemoryService()
    return MemULangGraphTools(service)


async def process_conversation(tools: list[BaseTool], user_id: str) -> None:
    """Simulate a conversation where memory is saved."""
    save_tool = next(t for t in tools if t.name == "save_memory")

    logger.info("--- Simulating Save Memory ---")
    inputs = {
        "content": "The user prefers dark mode and likes Python programming.",
        "user_id": user_id,
        "metadata": {"source": "demo_script"},
    }
    # Invoke the tool (async execution)
    result = await save_tool.ainvoke(inputs)
    logger.info("Save Result: %s", result)


async def process_retrieval(tools: list[BaseTool], user_id: str) -> None:
    """Simulate retrieving memory."""
    search_tool = next(t for t in tools if t.name == "search_memory")

    logger.info("--- Simulating Search Memory ---")
    inputs = {"query": "What are the user's preferences?", "user_id": user_id, "limit": 3}
    result = await search_tool.ainvoke(inputs)
    logger.info("Search Result:\n%s", result)


async def main() -> None:
    """Main entry point."""
    logger.info("Starting LangGraph Demo...")

    adapter = await initialize_infrastructure()
    tools = adapter.tools()

    user_id = "demo_user_123"

    await process_conversation(tools, user_id)
    await process_retrieval(tools, user_id)

    logger.info("Demo completed.")


if __name__ == "__main__":
    asyncio.run(main())



"""SQLite repository implementations for MemU."""

from memu.database.sqlite.repositories.base import SQLiteRepoBase
from memu.database.sqlite.repositories.category_item_repo import SQLiteCategoryItemRepo
from memu.database.sqlite.repositories.memory_category_repo import SQLiteMemoryCategoryRepo
from memu.database.sqlite.repositories.memory_item_repo import SQLiteMemoryItemRepo
from memu.database.sqlite.repositories.resource_repo import SQLiteResourceRepo

__all__ = [
    "SQLiteCategoryItemRepo",
    "SQLiteMemoryCategoryRepo",
    "SQLiteMemoryItemRepo",
    "SQLiteRepoBase",
    "SQLiteResourceRepo",
]



import json
from collections.abc import Awaitable
from pathlib import Path
from typing import Any

import pendulum

from .common import get_memory_service

USER_ID = "claude_user"


def dump_conversation_resource(
    conversation_messages: list[dict[str, Any]],
) -> str:
    resource_data = {
        "content": [
            {
                "role": message.get("role", "system"),
                "content": {"text": message.get("content", "")},
                "created_at": message.get("timestamp", pendulum.now().isoformat()),
            }
            for message in conversation_messages
        ]
    }
    time_string = pendulum.now().format("YYYYMMDD_HHmmss")
    resource_url = Path(__file__).parent / "data" / f"conv_{time_string}.json"
    resource_url.parent.mkdir(parents=True, exist_ok=True)
    with open(resource_url, "w") as f:
        json.dump(resource_data, f, indent=4, ensure_ascii=False)
    return resource_url.as_posix()


def memorize(conversation_messages: list[dict[str, Any]]) -> Awaitable[dict[str, Any]]:
    memory_service = get_memory_service()

    resource_url = dump_conversation_resource(conversation_messages)
    return memory_service.memorize(resource_url=resource_url, modality="conversation", user={"user_id": USER_ID})



PROMPT = """
# Task Objective
Analyze the provided document text and produce two outputs:
1. A condensed version that preserves all key information and important details while removing verbosity and redundancy.
2. A one-sentence caption summarizing what the document is about.

# Workflow
1. Read the **Document** (`{document_text}`) carefully to understand its full content.
2. Identify the main points, key arguments, and essential details.
3. Remove repetition, filler, and unnecessary verbosity while preserving meaning and completeness.
4. Rewrite the content in a concise, structured form.
5. Generate a single-sentence **Caption** that accurately summarizes the document's purpose or topic.

# Rules
- Preserve all key information, facts, and conclusions.
- Do not introduce new information or interpretations.
- Keep the processed content concise but complete.
- The caption must be exactly **one sentence**.
- Use only the information contained in the provided document.

# Output Format
Use the following structure:

<processed_content>
[Provide the condensed version of the document here]
</processed_content>

<caption>
[Provide a one-sentence summary of the document]
</caption>

# Input
Document:
{document_text}
"""



from __future__ import annotations

from typing import Any

from pydantic import BaseModel

from memu.database.inmemory.models import build_inmemory_models
from memu.database.inmemory.repositories import (
    InMemoryCategoryItemRepository,
    InMemoryMemoryCategoryRepository,
    InMemoryMemoryItemRepository,
    InMemoryResourceRepository,
)
from memu.database.inmemory.state import InMemoryState
from memu.database.interfaces import Database
from memu.database.models import CategoryItem, MemoryCategory, MemoryItem, Resource
from memu.database.repositories import MemoryCategoryRepo, ResourceRepo


class InMemoryStore(Database):
    def __init__(
        self,
        *,
        scope_model: type[BaseModel] | None = None,
        resource_model: type[Any] | None = None,
        memory_item_model: type[Any] | None = None,
        memory_category_model: type[Any] | None = None,
        category_item_model: type[Any] | None = None,
        state: InMemoryState | None = None,
    ) -> None:
        self.scope_model = scope_model or BaseModel
        (
            default_resource_model,
            default_memory_category_model,
            default_memory_item_model,
            default_category_item_model,
        ) = build_inmemory_models(self.scope_model)

        self.state = state or InMemoryState()
        self.resources: dict[str, Resource] = self.state.resources
        self.items: dict[str, MemoryItem] = self.state.items
        self.categories: dict[str, MemoryCategory] = self.state.categories
        self.relations: list[CategoryItem] = self.state.relations

        resource_model = resource_model or default_resource_model or Resource
        memory_item_model = memory_item_model or default_memory_item_model or MemoryItem
        memory_category_model = memory_category_model or default_memory_category_model or MemoryCategory
        category_item_model = category_item_model or default_category_item_model or CategoryItem

        self.resource_repo: ResourceRepo = InMemoryResourceRepository(state=self.state, resource_model=resource_model)
        self.memory_category_repo: MemoryCategoryRepo = InMemoryMemoryCategoryRepository(
            state=self.state, memory_category_model=memory_category_model
        )
        self.memory_item_repo = InMemoryMemoryItemRepository(state=self.state, memory_item_model=memory_item_model)
        self.category_item_repo = InMemoryCategoryItemRepository(
            state=self.state, category_item_model=category_item_model
        )

    def close(self) -> None:
        return None



import os

from memu.app import MemoryService

from ..config import memorize_config, retrieve_config

USER_ID = "claude_user"
SHARED_MEMORY_SERVICE = None


def get_memory_service() -> MemoryService:
    global SHARED_MEMORY_SERVICE
    if SHARED_MEMORY_SERVICE is not None:
        return SHARED_MEMORY_SERVICE

    api_key = os.getenv("OPENAI_API_KEY")
    if not api_key:
        msg = "Please set OPENAI_API_KEY environment variable"
        raise ValueError(msg)

    SHARED_MEMORY_SERVICE = MemoryService(
        llm_profiles={
            "default": {
                "api_key": api_key,
                "chat_model": "gpt-4o-mini",
            },
        },
        memorize_config=memorize_config,
        retrieve_config=retrieve_config,
    )
    return SHARED_MEMORY_SERVICE



from __future__ import annotations

import uuid
from collections.abc import Mapping
from typing import Any

import pendulum

from memu.database.inmemory.repositories.filter import matches_where
from memu.database.inmemory.state import InMemoryState
from memu.database.models import MemoryCategory
from memu.database.repositories.memory_category import MemoryCategoryRepo as MemoryCategoryRepoProtocol


class InMemoryMemoryCategoryRepository(MemoryCategoryRepoProtocol):
    def __init__(self, *, state: InMemoryState, memory_category_model: type[MemoryCategory]) -> None:
        self._state = state
        self.memory_category_model = memory_category_model
        self.categories: dict[str, MemoryCategory] = self._state.categories

    def list_categories(self, where: Mapping[str, Any] | None = None) -> dict[str, MemoryCategory]:
        if not where:
            return dict(self.categories)
        return {cid: cat for cid, cat in self.categories.items() if matches_where(cat, where)}

    def clear_categories(self, where: Mapping[str, Any] | None = None) -> dict[str, MemoryCategory]:
        if not where:
            matches = self.categories.copy()
            self.categories.clear()
            return matches
        matches = {cid: cat for cid, cat in self.categories.items() if matches_where(cat, where)}
        self.categories = {cid: cat for cid, cat in self.categories.items() if cid not in matches}
        return matches

    def get_or_create_category(
        self, *, name: str, description: str, embedding: list[float], user_data: dict[str, Any]
    ) -> MemoryCategory:
        for c in self.categories.values():
            if c.name == name and all(getattr(c, k) == v for k, v in user_data.items()):
                now = pendulum.now("UTC")
                if c.embedding is None:
                    c.embedding = embedding
                    c.updated_at = now
                if not c.description:
                    c.description = description
                    c.updated_at = now
                return c
        cid = str(uuid.uuid4())
        cat = self.memory_category_model(id=cid, name=name, description=description, embedding=embedding, **user_data)
        self.categories[cid] = cat
        return cat

    def update_category(
        self,
        *,
        category_id: str,
        name: str | None = None,
        description: str | None = None,
        embedding: list[float] | None = None,
        summary: str | None = None,
    ) -> MemoryCategory:
        cat = self.categories.get(category_id)
        if cat is None:
            msg = f"Category with id {category_id} not found"
            raise KeyError(msg)

        if name is not None:
            cat.name = name
        if description is not None:
            cat.description = description
        if embedding is not None:
            cat.embedding = embedding
        if summary is not None:
            cat.summary = summary

        cat.updated_at = pendulum.now("UTC")
        return cat

    def load_existing(self) -> None:
        return None


MemoryCategoryRepo = InMemoryMemoryCategoryRepository

__all__ = ["InMemoryMemoryCategoryRepository", "MemoryCategoryRepo"]



from __future__ import annotations

from typing import Any, cast

from memu.embedding.backends.base import EmbeddingBackend


class OpenAIEmbeddingBackend(EmbeddingBackend):
    """Backend for OpenAI-compatible embedding API."""

    name = "openai"
    embedding_endpoint = "/embeddings"

    def build_embedding_payload(self, *, inputs: list[str], embed_model: str) -> dict[str, Any]:
        return {"model": embed_model, "input": inputs}

    def parse_embedding_response(self, data: dict[str, Any]) -> list[list[float]]:
        return [cast(list[float], d["embedding"]) for d in data["data"]]



from __future__ import annotations

from pydantic import BaseModel

from memu.app.settings import DatabaseConfig
from memu.database.inmemory.models import build_inmemory_models
from memu.database.inmemory.repo import InMemoryStore


def build_inmemory_database(
    *,
    config: DatabaseConfig,
    user_model: type[BaseModel],
) -> InMemoryStore:
    resource_model, memory_category_model, memory_item_model, category_item_model = build_inmemory_models(user_model)
    return InMemoryStore(
        scope_model=user_model,
        resource_model=resource_model,
        memory_item_model=memory_item_model,
        memory_category_model=memory_category_model,
        category_item_model=category_item_model,
    )


__all__ = ["build_inmemory_database"]
